# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
import os

import nibabel as nb
import numpy as np

import pytest
from ...testing import utils
from ..confounds import CompCor, TCompCor, ACompCor


def close_up_to_column_sign(a, b, rtol=1e-05, atol=1e-08, equal_nan=False):
    """SVD can produce sign flips on a per-column basis."""
    kwargs = dict(rtol=rtol, atol=atol, equal_nan=equal_nan)
    if np.allclose(a, b, **kwargs):
        return True

    ret = True
    for acol, bcol in zip(a.T, b.T):
        ret &= np.allclose(acol, bcol, **kwargs) or np.allclose(acol, -bcol, **kwargs)
        if not ret:
            break

    return ret


@pytest.mark.parametrize(
    "a, b, close",
    [
        ([[0.1, 0.2], [0.3, 0.4]], [[-0.1, 0.2], [-0.3, 0.4]], True),
        ([[0.1, 0.2], [0.3, 0.4]], [[-0.1, 0.2], [0.3, -0.4]], False),
    ],
)
def test_close_up_to_column_sign(a, b, close):
    a = np.asanyarray(a)
    b = np.asanyarray(b)
    assert close_up_to_column_sign(a, b) == close
    # Sign flips of all columns never changes result
    assert close_up_to_column_sign(a, -b) == close
    assert close_up_to_column_sign(-a, b) == close
    assert close_up_to_column_sign(-a, -b) == close
    # Trivial case
    assert close_up_to_column_sign(a, a)
    assert close_up_to_column_sign(b, b)


class TestCompCor:
    """Note: Tests currently do a poor job of testing functionality"""

    filenames = {
        "functionalnii": "compcorfunc.nii",
        "masknii": "compcormask.nii",
        "masknii2": "compcormask2.nii",
        "components_file": None,
    }

    @pytest.fixture(autouse=True)
    def setup_class(self, tmpdir):
        # setup
        tmpdir.chdir()
        noise = np.fromfunction(self.fake_noise_fun, self.fake_data.shape)
        self.realigned_file = utils.save_toy_nii(
            self.fake_data + noise, self.filenames["functionalnii"]
        )
        mask = np.ones(self.fake_data.shape[:3])
        mask[0, 0, 0] = 0
        mask[0, 0, 1] = 0
        mask1 = utils.save_toy_nii(mask, self.filenames["masknii"])

        other_mask = np.ones(self.fake_data.shape[:3])
        other_mask[0, 1, 0] = 0
        other_mask[1, 1, 0] = 0
        mask2 = utils.save_toy_nii(other_mask, self.filenames["masknii2"])

        self.mask_files = [mask1, mask2]

    def test_compcor(self):
        expected_components = [
            [-0.1989607212, -0.5753813646],
            [0.5692369697, 0.5674945949],
            [-0.6662573243, 0.4675843432],
            [0.4206466244, -0.3361270124],
            [-0.1246655485, -0.1235705610],
        ]

        self.run_cc(
            CompCor(
                num_components=6,
                realigned_file=self.realigned_file,
                mask_files=self.mask_files,
                mask_index=0,
            ),
            expected_components,
        )

        self.run_cc(
            ACompCor(
                num_components=6,
                realigned_file=self.realigned_file,
                mask_files=self.mask_files,
                mask_index=0,
                components_file="acc_components_file",
            ),
            expected_components,
            "aCompCor",
        )

    def test_compcor_variance_threshold_and_metadata(self):
        expected_components = [
            [-0.2027150345, -0.4954813834],
            [0.2565929051, 0.7866217875],
            [-0.3550986008, -0.0089784905],
            [0.7512786244, -0.3599828482],
            [-0.4500578942, 0.0778209345],
        ]
        expected_metadata = {
            "component": "CompCor00",
            "mask": "mask",
            "singular_value": "4.0720553036",
            "variance_explained": "0.5527211465",
            "cumulative_variance_explained": "0.5527211465",
            "retained": "True",
        }
        ccinterface = CompCor(
            variance_threshold=0.7,
            realigned_file=self.realigned_file,
            mask_files=self.mask_files,
            mask_names=["mask"],
            mask_index=1,
            save_metadata=True,
        )
        self.run_cc(
            ccinterface=ccinterface,
            expected_components=expected_components,
            expected_n_components=2,
            expected_metadata=expected_metadata,
        )

    def test_tcompcor(self):
        ccinterface = TCompCor(
            num_components=6,
            realigned_file=self.realigned_file,
            percentile_threshold=0.75,
        )
        self.run_cc(
            ccinterface,
            [
                [-0.1114536190, -0.4632908609],
                [0.4566907310, 0.6983205193],
                [-0.7132557407, 0.1340170559],
                [0.5022537643, -0.5098322262],
                [-0.1342351356, 0.1407855119],
            ],
            "tCompCor",
        )

    def test_tcompcor_no_percentile(self):
        ccinterface = TCompCor(num_components=6, realigned_file=self.realigned_file)
        ccinterface.run()

        mask = nb.load("mask_000.nii.gz").dataobj
        num_nonmasked_voxels = np.count_nonzero(mask)
        assert num_nonmasked_voxels == 1

    def test_compcor_no_regress_poly(self):
        self.run_cc(
            CompCor(
                num_components=6,
                realigned_file=self.realigned_file,
                mask_files=self.mask_files,
                mask_index=0,
                pre_filter=False,
            ),
            [
                [0.4451946442, -0.7683311482],
                [-0.4285129505, -0.0926034137],
                [0.5721540256, 0.5608764842],
                [-0.5367548139, 0.0059943226],
                [-0.0520809054, 0.2940637551],
            ],
        )

    def test_tcompcor_asymmetric_dim(self):
        asymmetric_shape = (2, 3, 4, 5)
        asymmetric_data = utils.save_toy_nii(
            np.zeros(asymmetric_shape), "asymmetric.nii"
        )

        TCompCor(realigned_file=asymmetric_data).run()
        assert nb.load("mask_000.nii.gz").shape == asymmetric_shape[:3]

    def test_compcor_bad_input_shapes(self):
        # dim 0 is < dim 0 of self.mask_files (2)
        shape_less_than = (1, 2, 2, 5)
        # dim 0 is > dim 0 of self.mask_files (2)
        shape_more_than = (3, 3, 3, 5)

        for data_shape in (shape_less_than, shape_more_than):
            data_file = utils.save_toy_nii(np.zeros(data_shape), "temp.nii")
            interface = CompCor(realigned_file=data_file, mask_files=self.mask_files[0])
            with pytest.raises(ValueError):
                interface.run()  # Dimension mismatch

    def test_tcompcor_bad_input_dim(self):
        bad_dims = (2, 2, 2)
        data_file = utils.save_toy_nii(np.zeros(bad_dims), "temp.nii")
        interface = TCompCor(realigned_file=data_file)
        with pytest.raises(ValueError):
            interface.run()  # Not a 4D file

    def test_tcompcor_merge_intersect_masks(self):
        for method in ["union", "intersect"]:
            TCompCor(
                realigned_file=self.realigned_file,
                mask_files=self.mask_files,
                merge_method=method,
            ).run()
            if method == "union":
                assert np.array_equal(
                    nb.load("mask_000.nii.gz").dataobj,
                    ([[[0, 0], [0, 0]], [[0, 0], [1, 0]]]),
                )
            if method == "intersect":
                assert np.array_equal(
                    nb.load("mask_000.nii.gz").dataobj,
                    ([[[0, 0], [0, 0]], [[0, 1], [0, 0]]]),
                )

    def test_tcompcor_index_mask(self):
        TCompCor(
            realigned_file=self.realigned_file, mask_files=self.mask_files, mask_index=1
        ).run()
        assert np.array_equal(
            nb.load("mask_000.nii.gz").dataobj, ([[[0, 0], [0, 0]], [[0, 1], [0, 0]]])
        )

    def test_tcompcor_multi_mask_no_index(self):
        interface = TCompCor(
            realigned_file=self.realigned_file, mask_files=self.mask_files
        )
        with pytest.raises(ValueError):
            interface.run()  # more than one mask file

    def run_cc(
        self,
        ccinterface,
        expected_components,
        expected_header="CompCor",
        expected_n_components=None,
        expected_metadata=None,
    ):
        # run
        ccresult = ccinterface.run()

        # assert
        expected_file = ccinterface._list_outputs()["components_file"]
        assert ccresult.outputs.components_file == expected_file
        assert os.path.exists(expected_file)
        assert os.path.getsize(expected_file) > 0

        with open(ccresult.outputs.components_file, "r") as components_file:
            header = components_file.readline().rstrip().split("\t")
            components_data = np.loadtxt(components_file, delimiter="\t")

        if expected_n_components is None:
            expected_n_components = min(
                ccinterface.inputs.num_components, self.fake_data.shape[3]
            )

        assert header == [
            f"{expected_header}{i:02d}" for i in range(expected_n_components)
        ]

        assert components_data.shape == (self.fake_data.shape[3], expected_n_components)
        assert close_up_to_column_sign(components_data[:, :2], expected_components)

        if ccinterface.inputs.save_metadata:
            expected_metadata_file = ccinterface._list_outputs()["metadata_file"]
            assert ccresult.outputs.metadata_file == expected_metadata_file
            assert os.path.exists(expected_metadata_file)
            assert os.path.getsize(expected_metadata_file) > 0

            with open(ccresult.outputs.metadata_file, "r") as metadata_file:
                components_metadata = [
                    line.rstrip().split("\t") for line in metadata_file
                ]
                components_metadata = {
                    i: j for i, j in zip(components_metadata[0], components_metadata[1])
                }
                assert components_metadata == expected_metadata

        return ccresult

    @staticmethod
    def fake_noise_fun(i, j, l, m):
        return m * i + l - j

    fake_data = np.array(
        [
            [[[8, 5, 3, 8, 0], [6, 7, 4, 7, 1]], [[7, 9, 1, 6, 5], [0, 7, 4, 7, 7]]],
            [[[2, 4, 5, 7, 0], [1, 7, 0, 5, 4]], [[7, 3, 9, 0, 4], [9, 4, 1, 5, 0]]],
        ]
    )



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""The spm module provides basic functions for interfacing with SPM  tools.

In order to use the standalone MCR version of spm, you need to ensure that
the following commands are executed at the beginning of your script::

   from nipype.interfaces import spm
   matlab_cmd = '/path/to/run_spm8.sh /path/to/Compiler_Runtime/v713/ script'
   spm.SPMCommand.set_mlab_paths(matlab_cmd=matlab_cmd, use_mcr=True)

you can test by calling::

   spm.SPMCommand().version
"""
# Standard library imports
import os
from copy import deepcopy

# Third-party imports
from nibabel import load
import numpy as np

# Local imports
from ... import logging
from ...utils import spm_docs as sd
from ..base import (
    BaseInterface,
    traits,
    isdefined,
    InputMultiPath,
    BaseInterfaceInputSpec,
    Directory,
    Undefined,
    ImageFile,
    PackageInfo,
)
from ..base.traits_extension import NoDefaultSpecified
from ..matlab import MatlabCommand
from ...external.due import due, Doi, BibTeX

__docformat__ = "restructuredtext"
logger = logging.getLogger("nipype.interface")


def func_is_3d(in_file):
    """Checks if input functional files are 3d."""

    if isinstance(in_file, list):
        return func_is_3d(in_file[0])
    else:
        img = load(in_file)
        shape = img.shape
        if len(shape) == 3 or (len(shape) == 4 and shape[3] == 1):
            return True
        else:
            return False


def get_first_3dfile(in_files):
    if not func_is_3d(in_files):
        return None
    if isinstance(in_files[0], list):
        return in_files[0]
    return in_files


def scans_for_fname(fname):
    """Reads a nifti file and converts it to a numpy array storing
    individual nifti volumes.

    Opens images so will fail if they are not found.

    """
    if isinstance(fname, list):
        scans = np.zeros((len(fname),), dtype=object)
        for sno, f in enumerate(fname):
            scans[sno] = "%s,1" % f
        return scans
    img = load(fname)
    if len(img.shape) == 3:
        return np.array(("%s,1" % fname,), dtype=object)
    else:
        n_scans = img.shape[3]
        scans = np.zeros((n_scans,), dtype=object)
        for sno in range(n_scans):
            scans[sno] = "%s,%d" % (fname, sno + 1)
        return scans


def scans_for_fnames(fnames, keep4d=False, separate_sessions=False):
    """Converts a list of files to a concatenated numpy array for each
    volume.

    keep4d : boolean
        keeps the entries of the numpy array as 4d files instead of
        extracting the individual volumes.
    separate_sessions: boolean
        if 4d nifti files are being used, then separate_sessions
        ensures a cell array per session is created in the structure.

    """
    flist = None
    if not isinstance(fnames[0], list):
        if func_is_3d(fnames[0]):
            fnames = [fnames]
    if separate_sessions or keep4d:
        flist = np.zeros((len(fnames),), dtype=object)
    for i, f in enumerate(fnames):
        if separate_sessions:
            if keep4d:
                if isinstance(f, list):
                    flist[i] = np.array(f, dtype=object)
                else:
                    flist[i] = np.array([f], dtype=object)
            else:
                flist[i] = scans_for_fname(f)
        else:
            if keep4d:
                flist[i] = f
            else:
                scans = scans_for_fname(f)
                if flist is None:
                    flist = scans
                else:
                    flist = np.concatenate((flist, scans))
    return flist


class Info(PackageInfo):
    """Handles SPM version information

    If you use `SPMCommand.set_mlab_paths` to set alternate entries for
    matlab_cmd, paths, and use_mcr, then you will need to use the same entries
    to any call in the Info class to maintain memoization. Otherwise, it will
    default to the parameters in the `getinfo` function below.
    """

    _path = None
    _name = None
    _command = None
    _paths = None
    _version = None

    @classmethod
    def path(klass, matlab_cmd=None, paths=None, use_mcr=None):
        klass.getinfo(matlab_cmd, paths, use_mcr)
        return klass._path

    @classmethod
    def version(klass, matlab_cmd=None, paths=None, use_mcr=None):
        klass.getinfo(matlab_cmd, paths, use_mcr)
        return klass._version

    @classmethod
    def name(klass, matlab_cmd=None, paths=None, use_mcr=None):
        klass.getinfo(matlab_cmd, paths, use_mcr)
        return klass._name

    @classmethod
    def getinfo(klass, matlab_cmd=None, paths=None, use_mcr=None):
        """
        Returns the path to the SPM directory in the Matlab path
        If path not found, returns None.

        Parameters
        ----------
        matlab_cmd: str
            Sets the default matlab command. If None, the value of the
            environment variable SPMMCRCMD will be used if set and use_mcr
            is True or the environment variable FORCE_SPMMCR is set.
            If one of FORCE_SPMMCR or SPMMCRCMD is not set, the existence
            of the environment variable MATLABCMD is checked and its value
            is used as the matlab command if possible.
            If none of the above was successful, the fallback value of
            'matlab -nodesktop -nosplash' will be used.
        paths : str
            Add paths to matlab session
        use_mcr : bool
            Whether to use the MATLAB Common Runtime. In this case, the
            matlab_cmd is expected to be a valid MCR call.

        Returns
        -------
        spm_path : string representing path to SPM directory

            returns None of path not found
        """

        use_mcr = use_mcr or "FORCE_SPMMCR" in os.environ
        matlab_cmd = matlab_cmd or (
            (use_mcr and os.getenv("SPMMCRCMD"))
            or os.getenv("MATLABCMD", "matlab -nodesktop -nosplash")
        )

        if (
            klass._name
            and klass._path
            and klass._version
            and klass._command == matlab_cmd
            and klass._paths == paths
        ):

            return {"name": klass._name, "path": klass._path, "release": klass._version}
        logger.debug("matlab command or path has changed. recomputing version.")
        mlab = MatlabCommand(matlab_cmd=matlab_cmd, resource_monitor=False)
        mlab.inputs.mfile = False
        if paths:
            mlab.inputs.paths = paths
        if use_mcr:
            mlab.inputs.nodesktop = Undefined
            mlab.inputs.nosplash = Undefined
            mlab.inputs.single_comp_thread = Undefined
            mlab.inputs.mfile = True
            mlab.inputs.uses_mcr = True
        mlab.inputs.script = """
if isempty(which('spm')),
throw(MException('SPMCheck:NotFound','SPM not in matlab path'));
end;
spm_path = spm('dir');
[name, version] = spm('ver');
fprintf(1, 'NIPYPE path:%s|name:%s|release:%s', spm_path, name, version);
exit;
        """
        try:
            out = mlab.run()
        except (IOError, RuntimeError) as e:
            # if no Matlab at all -- exception could be raised
            # No Matlab -- no spm
            logger.debug("%s", e)
            klass._version = None
            klass._path = None
            klass._name = None
            klass._command = matlab_cmd
            klass._paths = paths
            return None

        out = sd._strip_header(out.runtime.stdout)
        out_dict = {}
        for part in out.split("|"):
            key, val = part.split(":")
            out_dict[key] = val

        klass._version = out_dict["release"]
        klass._path = out_dict["path"]
        klass._name = out_dict["name"]
        klass._command = matlab_cmd
        klass._paths = paths
        return out_dict


def no_spm():
    """Checks if SPM is NOT installed
    used with pytest.mark.skipif decorator to skip tests
    that will fail if spm is not installed"""

    if "NIPYPE_NO_MATLAB" in os.environ or Info.version() is None:
        return True
    else:
        return False


class SPMCommandInputSpec(BaseInterfaceInputSpec):
    matlab_cmd = traits.Str(desc="matlab command to use")
    paths = InputMultiPath(Directory(), desc="Paths to add to matlabpath")
    mfile = traits.Bool(True, desc="Run m-code using m-file", usedefault=True)
    use_mcr = traits.Bool(desc="Run m-code using SPM MCR")
    use_v8struct = traits.Bool(
        True,
        min_ver="8",
        usedefault=True,
        desc=("Generate SPM8 and higher " "compatible jobs"),
    )


class SPMCommand(BaseInterface):
    """Extends `BaseInterface` class to implement SPM specific interfaces.

    WARNING: Pseudo prototype class, meant to be subclassed
    """

    input_spec = SPMCommandInputSpec
    _additional_metadata = ["field"]

    _jobtype = "basetype"
    _jobname = "basename"

    _matlab_cmd = None
    _paths = None
    _use_mcr = None

    _references = [
        {
            "entry": BibTeX(
                "@book{FrackowiakFristonFrithDolanMazziotta1997,"
                "author={R.S.J. Frackowiak, K.J. Friston, C.D. Frith, R.J. Dolan, and J.C. Mazziotta},"
                "title={Human Brain Function},"
                "publisher={Academic Press USA},"
                "year={1997},"
                "}"
            ),
            "description": "The fundamental text on Statistical Parametric Mapping (SPM)",
            # 'path': "nipype.interfaces.spm",
            "tags": ["implementation"],
        }
    ]

    def __init__(self, **inputs):
        super(SPMCommand, self).__init__(**inputs)
        self.inputs.on_trait_change(
            self._matlab_cmd_update, ["matlab_cmd", "mfile", "paths", "use_mcr"]
        )
        self._find_mlab_cmd_defaults()
        self._check_mlab_inputs()
        self._matlab_cmd_update()

    @classmethod
    def set_mlab_paths(cls, matlab_cmd=None, paths=None, use_mcr=None):
        cls._matlab_cmd = matlab_cmd
        cls._paths = paths
        cls._use_mcr = use_mcr
        info_dict = Info.getinfo(matlab_cmd=matlab_cmd, paths=paths, use_mcr=use_mcr)

    def _find_mlab_cmd_defaults(self):
        # check if the user has set environment variables to enforce
        # the standalone (MCR) version of SPM
        if self._use_mcr or "FORCE_SPMMCR" in os.environ:
            self._use_mcr = True
            if self._matlab_cmd is None:
                try:
                    self._matlab_cmd = os.environ["SPMMCRCMD"]
                except KeyError:
                    pass

    def _matlab_cmd_update(self):
        # MatlabCommand has to be created here,
        # because matlab_cmd is not a proper input
        # and can be set only during init
        self.mlab = MatlabCommand(
            matlab_cmd=self.inputs.matlab_cmd,
            mfile=self.inputs.mfile,
            paths=self.inputs.paths,
            resource_monitor=False,
        )
        self.mlab.inputs.script_file = (
            "pyscript_%s.m" % self.__class__.__name__.split(".")[-1].lower()
        )
        if isdefined(self.inputs.use_mcr) and self.inputs.use_mcr:
            self.mlab.inputs.nodesktop = Undefined
            self.mlab.inputs.nosplash = Undefined
            self.mlab.inputs.single_comp_thread = Undefined
            self.mlab.inputs.uses_mcr = True
            self.mlab.inputs.mfile = True

    @property
    def version(self):
        info_dict = Info.getinfo(
            matlab_cmd=self.inputs.matlab_cmd,
            paths=self.inputs.paths,
            use_mcr=self.inputs.use_mcr,
        )
        if info_dict:
            return "%s.%s" % (info_dict["name"].split("SPM")[-1], info_dict["release"])

    @property
    def jobtype(self):
        return self._jobtype

    @property
    def jobname(self):
        return self._jobname

    def _check_mlab_inputs(self):
        if not isdefined(self.inputs.matlab_cmd) and self._matlab_cmd:
            self.inputs.matlab_cmd = self._matlab_cmd
        if not isdefined(self.inputs.paths) and self._paths:
            self.inputs.paths = self._paths
        if not isdefined(self.inputs.use_mcr) and self._use_mcr:
            self.inputs.use_mcr = self._use_mcr

    def _run_interface(self, runtime):
        """Executes the SPM function using MATLAB."""
        self.mlab.inputs.script = self._make_matlab_command(
            deepcopy(self._parse_inputs())
        )
        results = self.mlab.run()
        runtime.returncode = results.runtime.returncode
        if self.mlab.inputs.uses_mcr:
            if "Skipped" in results.runtime.stdout:
                self.raise_exception(runtime)
        runtime.stdout = results.runtime.stdout
        runtime.stderr = results.runtime.stderr
        runtime.merged = results.runtime.merged
        return runtime

    def _list_outputs(self):
        """Determine the expected outputs based on inputs."""

        raise NotImplementedError

    def _format_arg(self, opt, spec, val):
        """Convert input to appropriate format for SPM."""
        if spec.is_trait_type(traits.Bool):
            return int(val)
        elif spec.is_trait_type(traits.Tuple):
            return list(val)
        else:
            return val

    def _parse_inputs(self, skip=()):
        spmdict = {}
        metadata = dict(field=lambda t: t is not None)
        for name, spec in list(self.inputs.traits(**metadata).items()):
            if skip and name in skip:
                continue
            value = getattr(self.inputs, name)
            if not isdefined(value):
                continue
            field = spec.field
            if "." in field:
                fields = field.split(".")
                dictref = spmdict
                for f in fields[:-1]:
                    if f not in list(dictref.keys()):
                        dictref[f] = {}
                    dictref = dictref[f]
                dictref[fields[-1]] = self._format_arg(name, spec, value)
            else:
                spmdict[field] = self._format_arg(name, spec, value)
        return [spmdict]

    def _reformat_dict_for_savemat(self, contents):
        """Encloses a dict representation within hierarchical lists.

        In order to create an appropriate SPM job structure, a Python
        dict storing the job needs to be modified so that each dict
        embedded in dict needs to be enclosed as a list element.

        Examples
        --------
        >>> a = SPMCommand()._reformat_dict_for_savemat(dict(a=1,
        ...                                                  b=dict(c=2, d=3)))
        >>> a == [{'a': 1, 'b': [{'c': 2, 'd': 3}]}]
        True

        """
        newdict = {}
        try:
            for key, value in list(contents.items()):
                if isinstance(value, dict):
                    if value:
                        newdict[key] = self._reformat_dict_for_savemat(value)
                    # if value is None, skip
                else:
                    newdict[key] = value

            return [newdict]
        except TypeError:
            print("Requires dict input")

    def _generate_job(self, prefix="", contents=None):
        """Recursive function to generate spm job specification as a string

        Parameters
        ----------
        prefix : string
            A string that needs to get
        contents : dict
            A non-tuple Python structure containing spm job
            information gets converted to an appropriate sequence of
            matlab commands.

        """
        jobstring = ""
        if contents is None:
            return jobstring
        if isinstance(contents, list):
            for i, value in enumerate(contents):
                if prefix.endswith(")"):
                    newprefix = "%s,%d)" % (prefix[:-1], i + 1)
                else:
                    newprefix = "%s(%d)" % (prefix, i + 1)
                jobstring += self._generate_job(newprefix, value)
            return jobstring
        if isinstance(contents, dict):
            for key, value in list(contents.items()):
                newprefix = "%s.%s" % (prefix, key)
                jobstring += self._generate_job(newprefix, value)
            return jobstring
        if isinstance(contents, np.ndarray):
            if contents.dtype == np.dtype(object):
                if prefix:
                    jobstring += "%s = {...\n" % (prefix)
                else:
                    jobstring += "{...\n"
                for i, val in enumerate(contents):
                    if isinstance(val, np.ndarray):
                        jobstring += self._generate_job(prefix=None, contents=val)
                    elif isinstance(val, list):
                        items_format = []
                        for el in val:
                            items_format += [
                                "{}" if not isinstance(el, (str, bytes)) else "'{}'"
                            ]
                        val_format = ", ".join(items_format).format
                        jobstring += "[{}];...\n".format(val_format(*val))
                    elif isinstance(val, (str, bytes)):
                        jobstring += "'{}';...\n".format(val)
                    else:
                        jobstring += "%s;...\n" % str(val)
                jobstring += "};\n"
            else:
                for i, val in enumerate(contents):
                    for field in val.dtype.fields:
                        if prefix:
                            newprefix = "%s(%d).%s" % (prefix, i + 1, field)
                        else:
                            newprefix = "(%d).%s" % (i + 1, field)
                        jobstring += self._generate_job(newprefix, val[field])
            return jobstring
        if isinstance(contents, (str, bytes)):
            jobstring += "%s = '%s';\n" % (prefix, contents)
            return jobstring
        jobstring += "%s = %s;\n" % (prefix, str(contents))
        return jobstring

    def _make_matlab_command(self, contents, postscript=None):
        """Generates a mfile to build job structure
        Parameters
        ----------

        contents : list
            a list of dicts generated by _parse_inputs
            in each subclass

        cwd : string
            default os.getcwd()

        Returns
        -------
        mscript : string
            contents of a script called by matlab

        """
        cwd = os.getcwd()
        mscript = """
        %% Generated by nipype.interfaces.spm
        if isempty(which('spm')),
             throw(MException('SPMCheck:NotFound', 'SPM not in matlab path'));
        end
        [name, version] = spm('ver');
        fprintf('SPM version: %s Release: %s\\n',name, version);
        fprintf('SPM path: %s\\n', which('spm'));
        spm('Defaults','fMRI');

        if strcmp(name, 'SPM8') || strcmp(name(1:5), 'SPM12'),
           spm_jobman('initcfg');
           spm_get_defaults('cmdline', 1);
        end\n
        """
        if self.mlab.inputs.mfile:
            if isdefined(self.inputs.use_v8struct) and self.inputs.use_v8struct:
                mscript += self._generate_job(
                    "jobs{1}.spm.%s.%s" % (self.jobtype, self.jobname), contents[0]
                )
            else:
                if self.jobname in [
                    "st",
                    "smooth",
                    "preproc",
                    "preproc8",
                    "fmri_spec",
                    "fmri_est",
                    "factorial_design",
                    "defs",
                ]:
                    # parentheses
                    mscript += self._generate_job(
                        "jobs{1}.%s{1}.%s(1)" % (self.jobtype, self.jobname),
                        contents[0],
                    )
                else:
                    # curly brackets
                    mscript += self._generate_job(
                        "jobs{1}.%s{1}.%s{1}" % (self.jobtype, self.jobname),
                        contents[0],
                    )
        else:
            from scipy.io import savemat

            jobdef = {
                "jobs": [
                    {
                        self.jobtype: [
                            {self.jobname: self.reformat_dict_for_savemat(contents[0])}
                        ]
                    }
                ]
            }
            savemat(os.path.join(cwd, "pyjobs_%s.mat" % self.jobname), jobdef)
            mscript += "load pyjobs_%s;\n\n" % self.jobname
        mscript += """
        spm_jobman(\'run\', jobs);\n
        """
        if self.inputs.use_mcr:
            mscript += """
        if strcmp(name, 'SPM8') || strcmp(name(1:5), 'SPM12'),
            close(\'all\', \'force\');
        end;
            """
        if postscript is not None:
            mscript += postscript
        return mscript


class ImageFileSPM(ImageFile):
    """Defines a trait whose value must be a NIfTI file."""

    def __init__(
        self, value=NoDefaultSpecified, exists=False, resolve=False, **metadata
    ):
        """Create an ImageFileSPM trait."""
        super(ImageFileSPM, self).__init__(
            value=value,
            exists=exists,
            types=["nifti1", "nifti2"],
            allow_compressed=False,
            resolve=resolve,
            **metadata
        )



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:

import logging
from warnings import warn
import os
import sys
from .misc import str2bool

try:
    from ..external.cloghandler import ConcurrentRotatingFileHandler as RFHandler
except ImportError:
    # Next 2 lines are optional:  issue a warning to the user
    warn("ConcurrentLogHandler not installed. Using builtin log handler")
    from logging.handlers import RotatingFileHandler as RFHandler


class Logging(object):
    """Nipype logging class"""

    fmt = "%(asctime)s,%(msecs)d %(name)-2s " "%(levelname)-2s:\n\t %(message)s"
    datefmt = "%y%m%d-%H:%M:%S"

    def __init__(self, config):
        self._config = config
        # scope our logger to not interfere with user
        _nipype_logger = logging.getLogger("nipype")
        _nipype_hdlr = logging.StreamHandler(stream=sys.stdout)
        _nipype_hdlr.setFormatter(logging.Formatter(fmt=self.fmt, datefmt=self.datefmt))
        # if StreamHandler was added, do not stack
        if not len(_nipype_logger.handlers):
            _nipype_logger.addHandler(_nipype_hdlr)

        self._logger = logging.getLogger("nipype.workflow")
        self._utlogger = logging.getLogger("nipype.utils")
        self._fmlogger = logging.getLogger("nipype.filemanip")
        self._iflogger = logging.getLogger("nipype.interface")

        self.loggers = {
            "nipype.workflow": self._logger,
            "nipype.utils": self._utlogger,
            "nipype.filemanip": self._fmlogger,
            "nipype.interface": self._iflogger,
        }
        self._hdlr = None
        self.update_logging(self._config)

    def enable_file_logging(self):
        config = self._config
        LOG_FILENAME = os.path.join(
            config.get("logging", "log_directory"), "pypeline.log"
        )
        hdlr = RFHandler(
            LOG_FILENAME,
            maxBytes=int(config.get("logging", "log_size")),
            backupCount=int(config.get("logging", "log_rotate")),
        )
        formatter = logging.Formatter(fmt=self.fmt, datefmt=self.datefmt)
        hdlr.setFormatter(formatter)
        self._logger.addHandler(hdlr)
        self._utlogger.addHandler(hdlr)
        self._iflogger.addHandler(hdlr)
        self._fmlogger.addHandler(hdlr)
        self._hdlr = hdlr

    def disable_file_logging(self):
        if self._hdlr:
            self._logger.removeHandler(self._hdlr)
            self._utlogger.removeHandler(self._hdlr)
            self._iflogger.removeHandler(self._hdlr)
            self._fmlogger.removeHandler(self._hdlr)
            self._hdlr = None

    def update_logging(self, config):
        self._config = config
        self.disable_file_logging()
        self._logger.setLevel(
            logging.getLevelName(config.get("logging", "workflow_level"))
        )
        self._utlogger.setLevel(
            logging.getLevelName(config.get("logging", "utils_level"))
        )
        self._iflogger.setLevel(
            logging.getLevelName(config.get("logging", "interface_level"))
        )
        if str2bool(config.get("logging", "log_to_file")):
            self.enable_file_logging()

    def getLogger(self, name):
        if name == "filemanip":
            warn(
                'The "filemanip" logger has been deprecated and replaced by '
                'the "utils" logger as of nipype 1.0'
            )
        if name in self.loggers:
            return self.loggers[name]
        return None

    def getLevelName(self, name):
        return logging.getLevelName(name)

    def logdebug_dict_differences(self, dold, dnew, prefix=""):
        """Helper to log what actually changed from old to new values of
        dictionaries.

        typical use -- log difference for hashed_inputs
        """
        from .misc import dict_diff

        self._logger.warning(
            "logdebug_dict_differences has been deprecated, please use "
            "nipype.utils.misc.dict_diff."
        )
        self._logger.debug(dict_diff(dold, dnew))



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..model import SegStatsReconAll


def test_SegStatsReconAll_inputs():
    input_map = dict(
        annot=dict(
            argstr="--annot %s %s %s",
            mandatory=True,
            xor=("segmentation_file", "annot", "surf_label"),
        ),
        args=dict(
            argstr="%s",
        ),
        aseg=dict(
            extensions=None,
        ),
        avgwf_file=dict(
            argstr="--avgwfvol %s",
        ),
        avgwf_txt_file=dict(
            argstr="--avgwf %s",
        ),
        brain_vol=dict(
            argstr="--%s",
        ),
        brainmask_file=dict(
            argstr="--brainmask %s",
            extensions=None,
        ),
        calc_power=dict(
            argstr="--%s",
        ),
        calc_snr=dict(
            argstr="--snr",
        ),
        color_table_file=dict(
            argstr="--ctab %s",
            extensions=None,
            xor=("color_table_file", "default_color_table", "gca_color_table"),
        ),
        copy_inputs=dict(),
        cortex_vol_from_surf=dict(
            argstr="--surf-ctx-vol",
        ),
        default_color_table=dict(
            argstr="--ctab-default",
            xor=("color_table_file", "default_color_table", "gca_color_table"),
        ),
        empty=dict(
            argstr="--empty",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        etiv=dict(
            argstr="--etiv",
        ),
        etiv_only=dict(),
        euler=dict(
            argstr="--euler",
        ),
        exclude_ctx_gm_wm=dict(
            argstr="--excl-ctxgmwm",
        ),
        exclude_id=dict(
            argstr="--excludeid %d",
        ),
        frame=dict(
            argstr="--frame %d",
        ),
        gca_color_table=dict(
            argstr="--ctab-gca %s",
            extensions=None,
            xor=("color_table_file", "default_color_table", "gca_color_table"),
        ),
        in_file=dict(
            argstr="--i %s",
            extensions=None,
        ),
        in_intensity=dict(
            argstr="--in %s --in-intensity-name %s",
            extensions=None,
        ),
        intensity_units=dict(
            argstr="--in-intensity-units %s",
            requires=["in_intensity"],
        ),
        lh_orig_nofix=dict(
            extensions=None,
            mandatory=True,
        ),
        lh_pial=dict(
            extensions=None,
            mandatory=True,
        ),
        lh_white=dict(
            extensions=None,
            mandatory=True,
        ),
        mask_erode=dict(
            argstr="--maskerode %d",
        ),
        mask_file=dict(
            argstr="--mask %s",
            extensions=None,
        ),
        mask_frame=dict(
            requires=["mask_file"],
        ),
        mask_invert=dict(
            argstr="--maskinvert",
        ),
        mask_sign=dict(),
        mask_thresh=dict(
            argstr="--maskthresh %f",
        ),
        multiply=dict(
            argstr="--mul %f",
        ),
        non_empty_only=dict(
            argstr="--nonempty",
        ),
        partial_volume_file=dict(
            argstr="--pv %s",
            extensions=None,
        ),
        presurf_seg=dict(
            extensions=None,
        ),
        rh_orig_nofix=dict(
            extensions=None,
            mandatory=True,
        ),
        rh_pial=dict(
            extensions=None,
            mandatory=True,
        ),
        rh_white=dict(
            extensions=None,
            mandatory=True,
        ),
        ribbon=dict(
            extensions=None,
            mandatory=True,
        ),
        segment_id=dict(
            argstr="--id %s...",
        ),
        segmentation_file=dict(
            argstr="--seg %s",
            extensions=None,
            mandatory=True,
            xor=("segmentation_file", "annot", "surf_label"),
        ),
        sf_avg_file=dict(
            argstr="--sfavg %s",
        ),
        subcort_gm=dict(
            argstr="--subcortgray",
        ),
        subject_id=dict(
            argstr="--subject %s",
            mandatory=True,
            usedefault=True,
        ),
        subjects_dir=dict(),
        summary_file=dict(
            argstr="--sum %s",
            extensions=None,
            genfile=True,
            position=-1,
        ),
        supratent=dict(
            argstr="--supratent",
        ),
        surf_label=dict(
            argstr="--slabel %s %s %s",
            mandatory=True,
            xor=("segmentation_file", "annot", "surf_label"),
        ),
        total_gray=dict(
            argstr="--totalgray",
        ),
        transform=dict(
            extensions=None,
            mandatory=True,
        ),
        vox=dict(
            argstr="--vox %s",
        ),
        wm_vol_from_surf=dict(
            argstr="--surf-wm-vol",
        ),
    )
    inputs = SegStatsReconAll.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_SegStatsReconAll_outputs():
    output_map = dict(
        avgwf_file=dict(
            extensions=None,
        ),
        avgwf_txt_file=dict(
            extensions=None,
        ),
        sf_avg_file=dict(
            extensions=None,
        ),
        summary_file=dict(
            extensions=None,
        ),
    )
    outputs = SegStatsReconAll.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..specialized import BRAINSABC


def test_BRAINSABC_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        atlasDefinition=dict(
            argstr="--atlasDefinition %s",
            extensions=None,
        ),
        atlasToSubjectInitialTransform=dict(
            argstr="--atlasToSubjectInitialTransform %s",
            hash_files=False,
        ),
        atlasToSubjectTransform=dict(
            argstr="--atlasToSubjectTransform %s",
            hash_files=False,
        ),
        atlasToSubjectTransformType=dict(
            argstr="--atlasToSubjectTransformType %s",
        ),
        atlasWarpingOff=dict(
            argstr="--atlasWarpingOff ",
        ),
        debuglevel=dict(
            argstr="--debuglevel %d",
        ),
        defaultSuffix=dict(
            argstr="--defaultSuffix %s",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        filterIteration=dict(
            argstr="--filterIteration %d",
        ),
        filterMethod=dict(
            argstr="--filterMethod %s",
        ),
        filterTimeStep=dict(
            argstr="--filterTimeStep %f",
        ),
        gridSize=dict(
            argstr="--gridSize %s",
            sep=",",
        ),
        implicitOutputs=dict(
            argstr="--implicitOutputs %s...",
            hash_files=False,
        ),
        inputVolumeTypes=dict(
            argstr="--inputVolumeTypes %s",
            sep=",",
        ),
        inputVolumes=dict(
            argstr="--inputVolumes %s...",
        ),
        interpolationMode=dict(
            argstr="--interpolationMode %s",
        ),
        maxBiasDegree=dict(
            argstr="--maxBiasDegree %d",
        ),
        maxIterations=dict(
            argstr="--maxIterations %d",
        ),
        medianFilterSize=dict(
            argstr="--medianFilterSize %s",
            sep=",",
        ),
        numberOfSubSamplesInEachPlugArea=dict(
            argstr="--numberOfSubSamplesInEachPlugArea %s",
            sep=",",
        ),
        numberOfThreads=dict(
            argstr="--numberOfThreads %d",
        ),
        outputDir=dict(
            argstr="--outputDir %s",
            hash_files=False,
        ),
        outputDirtyLabels=dict(
            argstr="--outputDirtyLabels %s",
            hash_files=False,
        ),
        outputFormat=dict(
            argstr="--outputFormat %s",
        ),
        outputLabels=dict(
            argstr="--outputLabels %s",
            hash_files=False,
        ),
        outputVolumes=dict(
            argstr="--outputVolumes %s...",
            hash_files=False,
        ),
        posteriorTemplate=dict(
            argstr="--posteriorTemplate %s",
        ),
        purePlugsThreshold=dict(
            argstr="--purePlugsThreshold %f",
        ),
        restoreState=dict(
            argstr="--restoreState %s",
            extensions=None,
        ),
        saveState=dict(
            argstr="--saveState %s",
            hash_files=False,
        ),
        subjectIntermodeTransformType=dict(
            argstr="--subjectIntermodeTransformType %s",
        ),
        useKNN=dict(
            argstr="--useKNN ",
        ),
        writeLess=dict(
            argstr="--writeLess ",
        ),
    )
    inputs = BRAINSABC.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_BRAINSABC_outputs():
    output_map = dict(
        atlasToSubjectInitialTransform=dict(
            extensions=None,
        ),
        atlasToSubjectTransform=dict(
            extensions=None,
        ),
        implicitOutputs=dict(),
        outputDir=dict(),
        outputDirtyLabels=dict(
            extensions=None,
        ),
        outputLabels=dict(
            extensions=None,
        ),
        outputVolumes=dict(),
        saveState=dict(
            extensions=None,
        ),
    )
    outputs = BRAINSABC.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..diffusion import dtiestim


def test_dtiestim_inputs():
    input_map = dict(
        B0=dict(
            argstr="--B0 %s",
            hash_files=False,
        ),
        B0_mask_output=dict(
            argstr="--B0_mask_output %s",
            hash_files=False,
        ),
        DTI_double=dict(
            argstr="--DTI_double ",
        ),
        args=dict(
            argstr="%s",
        ),
        bad_region_mask=dict(
            argstr="--bad_region_mask %s",
            extensions=None,
        ),
        brain_mask=dict(
            argstr="--brain_mask %s",
            extensions=None,
        ),
        correction=dict(
            argstr="--correction %s",
        ),
        defaultTensor=dict(
            argstr="--defaultTensor %s",
            sep=",",
        ),
        dwi_image=dict(
            argstr="--dwi_image %s",
            extensions=None,
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        idwi=dict(
            argstr="--idwi %s",
            hash_files=False,
        ),
        method=dict(
            argstr="--method %s",
        ),
        shiftNeg=dict(
            argstr="--shiftNeg ",
        ),
        shiftNegCoeff=dict(
            argstr="--shiftNegCoeff %f",
        ),
        sigma=dict(
            argstr="--sigma %f",
        ),
        step=dict(
            argstr="--step %f",
        ),
        tensor_output=dict(
            argstr="--tensor_output %s",
            hash_files=False,
        ),
        threshold=dict(
            argstr="--threshold %d",
        ),
        verbose=dict(
            argstr="--verbose ",
        ),
        weight_iterations=dict(
            argstr="--weight_iterations %d",
        ),
    )
    inputs = dtiestim.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_dtiestim_outputs():
    output_map = dict(
        B0=dict(
            extensions=None,
        ),
        B0_mask_output=dict(
            extensions=None,
        ),
        idwi=dict(
            extensions=None,
        ),
        tensor_output=dict(
            extensions=None,
        ),
    )
    outputs = dtiestim.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..utils import TensorMetrics


def test_TensorMetrics_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        component=dict(
            argstr="-num %s",
            sep=",",
            usedefault=True,
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        in_file=dict(
            argstr="%s",
            extensions=None,
            mandatory=True,
            position=-1,
        ),
        in_mask=dict(
            argstr="-mask %s",
            extensions=None,
        ),
        modulate=dict(
            argstr="-modulate %s",
        ),
        out_ad=dict(
            argstr="-ad %s",
            extensions=None,
        ),
        out_adc=dict(
            argstr="-adc %s",
            extensions=None,
        ),
        out_cl=dict(
            argstr="-cl %s",
            extensions=None,
        ),
        out_cp=dict(
            argstr="-cp %s",
            extensions=None,
        ),
        out_cs=dict(
            argstr="-cs %s",
            extensions=None,
        ),
        out_eval=dict(
            argstr="-value %s",
            extensions=None,
        ),
        out_evec=dict(
            argstr="-vector %s",
            extensions=None,
        ),
        out_fa=dict(
            argstr="-fa %s",
            extensions=None,
        ),
        out_rd=dict(
            argstr="-rd %s",
            extensions=None,
        ),
    )
    inputs = TensorMetrics.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_TensorMetrics_outputs():
    output_map = dict(
        out_ad=dict(
            extensions=None,
        ),
        out_adc=dict(
            extensions=None,
        ),
        out_cl=dict(
            extensions=None,
        ),
        out_cp=dict(
            extensions=None,
        ),
        out_cs=dict(
            extensions=None,
        ),
        out_eval=dict(
            extensions=None,
        ),
        out_evec=dict(
            extensions=None,
        ),
        out_fa=dict(
            extensions=None,
        ),
        out_rd=dict(
            extensions=None,
        ),
    )
    outputs = TensorMetrics.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""Miscellaneous utility functions
"""
import os
import sys
import gc
import errno
import select
import locale
import datetime
from pathlib import Path
from subprocess import Popen, STDOUT, PIPE
from .filemanip import canonicalize_env, read_stream

from .. import logging

iflogger = logging.getLogger("nipype.interface")


class Stream(object):
    """Function to capture stdout and stderr streams with timestamps

    stackoverflow.com/questions/4984549/merge-and-sync-stdout-and-stderr/5188359
    """

    def __init__(self, name, impl):
        self._name = name
        self._impl = impl
        self._buf = ""
        self._rows = []
        self._lastidx = 0
        self.default_encoding = locale.getdefaultlocale()[1] or "UTF-8"

    def fileno(self):
        "Pass-through for file descriptor."
        return self._impl.fileno()

    def read(self, drain=0):
        "Read from the file descriptor. If 'drain' set, read until EOF."
        while self._read(drain) is not None:
            if not drain:
                break

    def _read(self, drain):
        "Read from the file descriptor"
        fd = self.fileno()
        buf = os.read(fd, 4096).decode(self.default_encoding)
        if not buf and not self._buf:
            return None
        if "\n" not in buf:
            if not drain:
                self._buf += buf
                return []

        # prepend any data previously read, then split into lines and format
        buf = self._buf + buf
        if "\n" in buf:
            tmp, rest = buf.rsplit("\n", 1)
        else:
            tmp = buf
            rest = None
        self._buf = rest
        now = datetime.datetime.now().isoformat()
        rows = tmp.split("\n")
        self._rows += [(now, "%s %s:%s" % (self._name, now, r), r) for r in rows]
        for idx in range(self._lastidx, len(self._rows)):
            iflogger.info(self._rows[idx][1])
        self._lastidx = len(self._rows)


def run_command(runtime, output=None, timeout=0.01, write_cmdline=False):
    """Run a command, read stdout and stderr, prefix with timestamp.

    The returned runtime contains a merged stdout+stderr log with timestamps
    """

    # Init variables
    cmdline = runtime.cmdline
    env = canonicalize_env(runtime.environ)

    errfile = None
    outfile = None
    stdout = PIPE
    stderr = PIPE

    if output == "file":
        outfile = os.path.join(runtime.cwd, "output.nipype")
        stdout = open(outfile, "wb")  # t=='text'===default
        stderr = STDOUT
    elif output == "file_split":
        outfile = os.path.join(runtime.cwd, "stdout.nipype")
        stdout = open(outfile, "wb")
        errfile = os.path.join(runtime.cwd, "stderr.nipype")
        stderr = open(errfile, "wb")
    elif output == "file_stdout":
        outfile = os.path.join(runtime.cwd, "stdout.nipype")
        stdout = open(outfile, "wb")
    elif output == "file_stderr":
        errfile = os.path.join(runtime.cwd, "stderr.nipype")
        stderr = open(errfile, "wb")

    if write_cmdline:
        (Path(runtime.cwd) / "command.txt").write_text(cmdline)

    proc = Popen(
        cmdline,
        stdout=stdout,
        stderr=stderr,
        shell=True,
        cwd=runtime.cwd,
        env=env,
        close_fds=(not sys.platform.startswith("win")),
    )

    result = {
        "stdout": [],
        "stderr": [],
        "merged": [],
    }

    if output == "stream":
        streams = [Stream("stdout", proc.stdout), Stream("stderr", proc.stderr)]

        def _process(drain=0):
            try:
                res = select.select(streams, [], [], timeout)
            except select.error as e:
                iflogger.info(e)
                if e.errno == errno.EINTR:
                    return
                else:
                    raise
            else:
                for stream in res[0]:
                    stream.read(drain)

        while proc.returncode is None:
            proc.poll()
            _process()

        _process(drain=1)

        # collect results, merge and return
        result = {}
        temp = []
        for stream in streams:
            rows = stream._rows
            temp += rows
            result[stream._name] = [r[2] for r in rows]
        temp.sort()
        result["merged"] = [r[1] for r in temp]

    if output.startswith("file"):
        proc.wait()
        if outfile is not None:
            stdout.flush()
            stdout.close()
            with open(outfile, "rb") as ofh:
                stdoutstr = ofh.read()
            result["stdout"] = read_stream(stdoutstr, logger=iflogger)
            del stdoutstr

        if errfile is not None:
            stderr.flush()
            stderr.close()
            with open(errfile, "rb") as efh:
                stderrstr = efh.read()
            result["stderr"] = read_stream(stderrstr, logger=iflogger)
            del stderrstr

        if output == "file":
            result["merged"] = result["stdout"]
            result["stdout"] = []
    else:
        stdout, stderr = proc.communicate()
        if output == "allatonce":  # Discard stdout and stderr otherwise
            result["stdout"] = read_stream(stdout, logger=iflogger)
            result["stderr"] = read_stream(stderr, logger=iflogger)

    runtime.returncode = proc.returncode
    try:
        proc.terminate()  # Ensure we are done
    except OSError as error:
        # Python 2 raises when the process is already gone
        if error.errno != errno.ESRCH:
            raise

    # Dereference & force GC for a cleanup
    del proc
    del stdout
    del stderr
    gc.collect()

    runtime.stderr = "\n".join(result["stderr"])
    runtime.stdout = "\n".join(result["stdout"])
    runtime.merged = "\n".join(result["merged"])
    return runtime



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
Please write tests for all code submitted to the repository. The code will be
used by many people, and will in due course be used in live analyses, so we
need to make sure that we have the best possible defenses against bugs. It also
helps us think about code interfaces, and gives examples of code use that can
be useful for others using the code.

Python's unit testing framework (the
U{unittest<http://docs.python.org/lib/module-unittest.html>} module) is used to
implement project tests. We use the convention that each package contains a
subpackage called tests which contains modules defining test cases (subclasses
of U{unittest.TestCase<http://docs.python.org/lib/testcase-objects.html>})
for that package. The nipy.utils.tests package contains an example test case
called L{test_template.TemplateTest} to get you started writing your tests.
Please try to include working test cases for all functions and classes that
you contribute. Often, writing tests for your code before the code is written
helps to frame your thoughts about what the code should look like.
"""



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..model import TwoSampleTTestDesign


def test_TwoSampleTTestDesign_inputs():
    input_map = dict(
        covariates=dict(
            field="cov",
        ),
        dependent=dict(
            field="des.t2.dept",
        ),
        explicit_mask_file=dict(
            extensions=None,
            field="masking.em",
        ),
        global_calc_mean=dict(
            field="globalc.g_mean",
            xor=["global_calc_omit", "global_calc_values"],
        ),
        global_calc_omit=dict(
            field="globalc.g_omit",
            xor=["global_calc_mean", "global_calc_values"],
        ),
        global_calc_values=dict(
            field="globalc.g_user.global_uval",
            xor=["global_calc_mean", "global_calc_omit"],
        ),
        global_normalization=dict(
            field="globalm.glonorm",
        ),
        group1_files=dict(
            field="des.t2.scans1",
            mandatory=True,
        ),
        group2_files=dict(
            field="des.t2.scans2",
            mandatory=True,
        ),
        matlab_cmd=dict(),
        mfile=dict(
            usedefault=True,
        ),
        no_grand_mean_scaling=dict(
            field="globalm.gmsca.gmsca_no",
        ),
        paths=dict(),
        spm_mat_dir=dict(
            field="dir",
        ),
        threshold_mask_absolute=dict(
            field="masking.tm.tma.athresh",
            xor=["threshold_mask_none", "threshold_mask_relative"],
        ),
        threshold_mask_none=dict(
            field="masking.tm.tm_none",
            xor=["threshold_mask_absolute", "threshold_mask_relative"],
        ),
        threshold_mask_relative=dict(
            field="masking.tm.tmr.rthresh",
            xor=["threshold_mask_absolute", "threshold_mask_none"],
        ),
        unequal_variance=dict(
            field="des.t2.variance",
        ),
        use_implicit_threshold=dict(
            field="masking.im",
        ),
        use_mcr=dict(),
        use_v8struct=dict(
            min_ver="8",
            usedefault=True,
        ),
    )
    inputs = TwoSampleTTestDesign.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_TwoSampleTTestDesign_outputs():
    output_map = dict(
        spm_mat_file=dict(
            extensions=None,
        ),
    )
    outputs = TwoSampleTTestDesign.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
# -*- coding: utf-8 -*-
"""MRTrix3 provides software tools to perform various types of diffusion MRI analyses."""
from .connectivity import BuildConnectome, LabelConfig, LabelConvert
from .preprocess import (
    ACTPrepareFSL,
    DWIBiasCorrect,
    DWIDenoise,
    DWIPreproc,
    MRDeGibbs,
    ReplaceFSwithFIRST,
    ResponseSD,
)
from .reconst import ConstrainedSphericalDeconvolution, EstimateFOD, FitTensor
from .tracking import Tractography
from .utils import (
    TCK2VTK,
    BrainMask,
    ComputeTDI,
    DWIExtract,
    Generate5tt,
    Mesh2PVE,
    MRCat,
    MRConvert,
    MRMath,
    MRResize,
    MRTransform,
    SH2Amp,
    SHConv,
    TensorMetrics,
    TransformFSLConvert,
)



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..io import XNATSink


def test_XNATSink_inputs():
    input_map = dict(
        _outputs=dict(
            usedefault=True,
        ),
        assessor_id=dict(
            xor=["reconstruction_id"],
        ),
        cache_dir=dict(),
        config=dict(
            extensions=None,
            mandatory=True,
            xor=["server"],
        ),
        experiment_id=dict(
            mandatory=True,
        ),
        project_id=dict(
            mandatory=True,
        ),
        pwd=dict(),
        reconstruction_id=dict(
            xor=["assessor_id"],
        ),
        server=dict(
            mandatory=True,
            requires=["user", "pwd"],
            xor=["config"],
        ),
        share=dict(
            usedefault=True,
        ),
        subject_id=dict(
            mandatory=True,
        ),
        user=dict(),
    )
    inputs = XNATSink.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..brainsuite import BDP


def test_BDP_inputs():
    input_map = dict(
        BVecBValPair=dict(
            argstr="--bvec %s --bval %s",
            mandatory=True,
            position=-1,
            xor=["bMatrixFile"],
        ),
        args=dict(
            argstr="%s",
        ),
        bMatrixFile=dict(
            argstr="--bmat %s",
            extensions=None,
            mandatory=True,
            position=-1,
            xor=["BVecBValPair"],
        ),
        bValRatioThreshold=dict(
            argstr="--bval-ratio-threshold %f",
        ),
        bfcFile=dict(
            argstr="%s",
            extensions=None,
            mandatory=True,
            position=0,
            xor=["noStructuralRegistration"],
        ),
        customDiffusionLabel=dict(
            argstr="--custom-diffusion-label %s",
            extensions=None,
        ),
        customLabelXML=dict(
            argstr="--custom-label-xml %s",
            extensions=None,
        ),
        customT1Label=dict(
            argstr="--custom-t1-label %s",
            extensions=None,
        ),
        dataSinkDelay=dict(
            argstr="%s",
        ),
        dcorrRegMeasure=dict(
            argstr="--dcorr-reg-method %s",
        ),
        dcorrWeight=dict(
            argstr="--dcorr-regularization-wt %f",
        ),
        dwiMask=dict(
            argstr="--dwi-mask %s",
            extensions=None,
        ),
        echoSpacing=dict(
            argstr="--echo-spacing=%f",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        estimateODF_3DShore=dict(
            argstr="--3dshore --diffusion_time_ms %f",
        ),
        estimateODF_FRACT=dict(
            argstr="--FRACT",
        ),
        estimateODF_FRT=dict(
            argstr="--FRT",
        ),
        estimateTensors=dict(
            argstr="--tensors",
        ),
        fieldmapCorrection=dict(
            argstr="--fieldmap-correction %s",
            extensions=None,
            requires=["echoSpacing"],
        ),
        fieldmapCorrectionMethod=dict(
            argstr="--fieldmap-correction-method %s",
            xor=["skipIntensityCorr"],
        ),
        fieldmapSmooth=dict(
            argstr="--fieldmap-smooth3=%f",
        ),
        flagConfigFile=dict(
            argstr="--flag-conf-file %s",
            extensions=None,
        ),
        forcePartialROIStats=dict(
            argstr="--force-partial-roi-stats",
        ),
        generateStats=dict(
            argstr="--generate-stats",
        ),
        ignoreFieldmapFOV=dict(
            argstr="--ignore-fieldmap-fov",
        ),
        ignoreMemory=dict(
            argstr="--ignore-memory",
        ),
        inputDiffusionData=dict(
            argstr="--nii %s",
            extensions=None,
            mandatory=True,
            position=-2,
        ),
        lowMemory=dict(
            argstr="--low-memory",
        ),
        noStructuralRegistration=dict(
            argstr="--no-structural-registration",
            mandatory=True,
            position=0,
            xor=["bfcFile"],
        ),
        odfLambta=dict(
            argstr="--odf-lambda <L>",
        ),
        onlyStats=dict(
            argstr="--generate-only-stats",
        ),
        outPrefix=dict(
            argstr="--output-fileprefix %s",
        ),
        outputDiffusionCoordinates=dict(
            argstr="--output-diffusion-coordinate",
        ),
        outputSubdir=dict(
            argstr="--output-subdir %s",
        ),
        phaseEncodingDirection=dict(
            argstr="--dir=%s",
        ),
        rigidRegMeasure=dict(
            argstr="--rigid-reg-measure %s",
        ),
        skipDistortionCorr=dict(
            argstr="--no-distortion-correction",
        ),
        skipIntensityCorr=dict(
            argstr="--no-intensity-correction",
            xor=["fieldmapCorrectionMethod"],
        ),
        skipNonuniformityCorr=dict(
            argstr="--no-nonuniformity-correction",
        ),
        t1Mask=dict(
            argstr="--t1-mask %s",
            extensions=None,
        ),
        threads=dict(
            argstr="--threads=%d",
        ),
        transformDataOnly=dict(
            argstr="--transform-data-only",
        ),
        transformDiffusionSurface=dict(
            argstr="--transform-diffusion-surface %s",
            extensions=None,
        ),
        transformDiffusionVolume=dict(
            argstr="--transform-diffusion-volume %s",
            extensions=None,
        ),
        transformInterpolation=dict(
            argstr="--transform-interpolation %s",
        ),
        transformT1Surface=dict(
            argstr="--transform-t1-surface %s",
            extensions=None,
        ),
        transformT1Volume=dict(
            argstr="--transform-t1-volume %s",
            extensions=None,
        ),
    )
    inputs = BDP.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
"""Parallel workflow execution via Condor
"""

import os
from time import sleep

from ...interfaces.base import CommandLine
from ... import logging
from .base import SGELikeBatchManagerBase, logger

iflogger = logging.getLogger("nipype.interface")


class CondorPlugin(SGELikeBatchManagerBase):
    """Execute using Condor

    This plugin doesn't work with a plain stock-Condor installation, but
    requires a 'qsub' emulation script for Condor, called 'condor_qsub'.
    This script is shipped with the Condor package from NeuroDebian, or can be
    downloaded from its Git repository at

    http://anonscm.debian.org/gitweb/?p=pkg-exppsy/condor.git;a=blob_plain;f=debian/condor_qsub;hb=HEAD

    The plugin_args input to run can be used to control the Condor execution.
    Currently supported options are:

    - template : template to use for batch job submission. This can be an
                 SGE-style script with the (limited) set of options supported
                 by condor_qsub
    - qsub_args : arguments to be prepended to the job execution script in the
                  qsub call
    """

    def __init__(self, **kwargs):
        template = """
#$ -V
#$ -S /bin/sh
        """
        self._retry_timeout = 2
        self._max_tries = 2
        if "plugin_args" in kwargs and kwargs["plugin_args"]:
            if "retry_timeout" in kwargs["plugin_args"]:
                self._retry_timeout = kwargs["plugin_args"]["retry_timeout"]
            if "max_tries" in kwargs["plugin_args"]:
                self._max_tries = kwargs["plugin_args"]["max_tries"]
        super(CondorPlugin, self).__init__(template, **kwargs)

    def _is_pending(self, taskid):
        cmd = CommandLine(
            "condor_q", resource_monitor=False, terminal_output="allatonce"
        )
        cmd.inputs.args = "%d" % taskid
        # check condor cluster
        oldlevel = iflogger.level
        iflogger.setLevel(logging.getLevelName("CRITICAL"))
        result = cmd.run(ignore_exception=True)
        iflogger.setLevel(oldlevel)
        if result.runtime.stdout.count("\n%d" % taskid):
            return True
        return False

    def _submit_batchtask(self, scriptfile, node):
        cmd = CommandLine(
            "condor_qsub",
            environ=dict(os.environ),
            resource_monitor=False,
            terminal_output="allatonce",
        )
        path = os.path.dirname(scriptfile)
        qsubargs = ""
        if self._qsub_args:
            qsubargs = self._qsub_args
        if "qsub_args" in node.plugin_args:
            if "overwrite" in node.plugin_args and node.plugin_args["overwrite"]:
                qsubargs = node.plugin_args["qsub_args"]
            else:
                qsubargs += " " + node.plugin_args["qsub_args"]
        if self._qsub_args:
            qsubargs = self._qsub_args
        if "-o" not in qsubargs:
            qsubargs = "%s -o %s" % (qsubargs, path)
        if "-e" not in qsubargs:
            qsubargs = "%s -e %s" % (qsubargs, path)
        if node._hierarchy:
            jobname = ".".join((dict(os.environ)["LOGNAME"], node._hierarchy, node._id))
        else:
            jobname = ".".join((dict(os.environ)["LOGNAME"], node._id))
        jobnameitems = jobname.split(".")
        jobnameitems.reverse()
        jobname = ".".join(jobnameitems)
        cmd.inputs.args = "%s -N %s %s" % (qsubargs, jobname, scriptfile)
        oldlevel = iflogger.level
        iflogger.setLevel(logging.getLevelName("CRITICAL"))
        tries = 0
        while True:
            try:
                result = cmd.run()
            except Exception as e:
                if tries < self._max_tries:
                    tries += 1
                    sleep(self._retry_timeout)  # sleep 2 seconds and try again
                else:
                    iflogger.setLevel(oldlevel)
                    raise RuntimeError(
                        "\n".join(
                            (
                                ("Could not submit condor " "cluster" " for node %s")
                                % node._id,
                                str(e),
                            )
                        )
                    )
            else:
                break
        iflogger.setLevel(oldlevel)
        # retrieve condor clusterid
        taskid = int(result.runtime.stdout.split(" ")[2])
        self._pending[taskid] = node.output_dir()
        logger.debug("submitted condor cluster: %d for node %s" % (taskid, node._id))
        return taskid



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..fibertrack import fibertrack


def test_fibertrack_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        forbidden_label=dict(
            argstr="--forbidden_label %d",
        ),
        force=dict(
            argstr="--force ",
        ),
        input_roi_file=dict(
            argstr="--input_roi_file %s",
            extensions=None,
        ),
        input_tensor_file=dict(
            argstr="--input_tensor_file %s",
            extensions=None,
        ),
        max_angle=dict(
            argstr="--max_angle %f",
        ),
        min_fa=dict(
            argstr="--min_fa %f",
        ),
        output_fiber_file=dict(
            argstr="--output_fiber_file %s",
            hash_files=False,
        ),
        really_verbose=dict(
            argstr="--really_verbose ",
        ),
        source_label=dict(
            argstr="--source_label %d",
        ),
        step_size=dict(
            argstr="--step_size %f",
        ),
        target_label=dict(
            argstr="--target_label %d",
        ),
        verbose=dict(
            argstr="--verbose ",
        ),
        whole_brain=dict(
            argstr="--whole_brain ",
        ),
    )
    inputs = fibertrack.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_fibertrack_outputs():
    output_map = dict(
        output_fiber_file=dict(
            extensions=None,
        ),
    )
    outputs = fibertrack.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..utils import SurfaceSnapshots


def test_SurfaceSnapshots_inputs():
    input_map = dict(
        annot_file=dict(
            argstr="-annotation %s",
            extensions=None,
            xor=["annot_name"],
        ),
        annot_name=dict(
            argstr="-annotation %s",
            xor=["annot_file"],
        ),
        args=dict(
            argstr="%s",
        ),
        colortable=dict(
            argstr="-colortable %s",
            extensions=None,
        ),
        demean_overlay=dict(
            argstr="-zm",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        hemi=dict(
            argstr="%s",
            mandatory=True,
            position=2,
        ),
        identity_reg=dict(
            argstr="-overlay-reg-identity",
            xor=["overlay_reg", "identity_reg", "mni152_reg"],
        ),
        invert_overlay=dict(
            argstr="-invphaseflag 1",
        ),
        label_file=dict(
            argstr="-label %s",
            extensions=None,
            xor=["label_name"],
        ),
        label_name=dict(
            argstr="-label %s",
            xor=["label_file"],
        ),
        label_outline=dict(
            argstr="-label-outline",
        ),
        label_under=dict(
            argstr="-labels-under",
        ),
        mni152_reg=dict(
            argstr="-mni152reg",
            xor=["overlay_reg", "identity_reg", "mni152_reg"],
        ),
        orig_suffix=dict(
            argstr="-orig %s",
        ),
        overlay=dict(
            argstr="-overlay %s",
            extensions=None,
            requires=["overlay_range"],
        ),
        overlay_range=dict(
            argstr="%s",
        ),
        overlay_range_offset=dict(
            argstr="-foffset %.3f",
        ),
        overlay_reg=dict(
            argstr="-overlay-reg %s",
            extensions=None,
            xor=["overlay_reg", "identity_reg", "mni152_reg"],
        ),
        patch_file=dict(
            argstr="-patch %s",
            extensions=None,
        ),
        reverse_overlay=dict(
            argstr="-revphaseflag 1",
        ),
        screenshot_stem=dict(),
        show_color_scale=dict(
            argstr="-colscalebarflag 1",
        ),
        show_color_text=dict(
            argstr="-colscaletext 1",
        ),
        show_curv=dict(
            argstr="-curv",
            xor=["show_gray_curv"],
        ),
        show_gray_curv=dict(
            argstr="-gray",
            xor=["show_curv"],
        ),
        six_images=dict(),
        sphere_suffix=dict(
            argstr="-sphere %s",
        ),
        stem_template_args=dict(
            requires=["screenshot_stem"],
        ),
        subject_id=dict(
            argstr="%s",
            mandatory=True,
            position=1,
        ),
        subjects_dir=dict(),
        surface=dict(
            argstr="%s",
            mandatory=True,
            position=3,
        ),
        tcl_script=dict(
            argstr="%s",
            extensions=None,
            genfile=True,
        ),
        truncate_overlay=dict(
            argstr="-truncphaseflag 1",
        ),
    )
    inputs = SurfaceSnapshots.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_SurfaceSnapshots_outputs():
    output_map = dict(
        snapshots=dict(),
    )
    outputs = SurfaceSnapshots.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..brainsuite import Tca


def test_Tca_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        foregroundDelta=dict(
            argstr="--delta %d",
            usedefault=True,
        ),
        inputMaskFile=dict(
            argstr="-i %s",
            extensions=None,
            mandatory=True,
        ),
        maxCorrectionSize=dict(
            argstr="-n %d",
        ),
        minCorrectionSize=dict(
            argstr="-m %d",
            usedefault=True,
        ),
        outputMaskFile=dict(
            argstr="-o %s",
            extensions=None,
            genfile=True,
        ),
        timer=dict(
            argstr="--timer",
        ),
        verbosity=dict(
            argstr="-v %d",
        ),
    )
    inputs = Tca.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_Tca_outputs():
    output_map = dict(
        outputMaskFile=dict(
            extensions=None,
        ),
    )
    outputs = Tca.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



from nipype.interfaces.ants.base import Info

import pytest

# fmt: off
ANTS_VERSIONS = [("""\
ANTs Version: 2.3.3.dev168-g29bdf
Compiled: Jun  9 2020 03:44:55

""", "2.3.3"), ("""\
ANTs Version: v2.3.5.post76-g28dd25c
Compiled: Nov 16 2021 14:57:48

""", "2.3.5"), ("""\
ANTs Version: v2.1.0.post789-g0740f
Compiled: I don't still have this so not going to pretend

""", "2.2.0"),
]
# fmt: on


@pytest.mark.parametrize("raw_info, version", ANTS_VERSIONS)
def test_version_parser(raw_info, version):
    assert Info.parse_version(raw_info) == version



# -*- coding: utf-8 -*-
from multiprocessing import Pool, cpu_count
import os.path as op

import numpy as np
import nibabel as nb

from ... import logging
from ..base import (
    traits,
    TraitedSpec,
    BaseInterfaceInputSpec,
    File,
    InputMultiPath,
    isdefined,
)
from .base import DipyBaseInterface

IFLOGGER = logging.getLogger("nipype.interface")


class SimulateMultiTensorInputSpec(BaseInterfaceInputSpec):
    in_dirs = InputMultiPath(
        File(exists=True), mandatory=True, desc="list of fibers (principal directions)"
    )
    in_frac = InputMultiPath(
        File(exists=True), mandatory=True, desc=("volume fraction of each fiber")
    )
    in_vfms = InputMultiPath(
        File(exists=True),
        mandatory=True,
        desc=("volume fractions of isotropic " "compartiments"),
    )
    in_mask = File(exists=True, desc="mask to simulate data")

    diff_iso = traits.List(
        [3000e-6, 960e-6, 680e-6],
        traits.Float,
        usedefault=True,
        desc="Diffusivity of isotropic compartments",
    )
    diff_sf = traits.Tuple(
        (1700e-6, 200e-6, 200e-6),
        traits.Float,
        traits.Float,
        traits.Float,
        usedefault=True,
        desc="Single fiber tensor",
    )

    n_proc = traits.Int(0, usedefault=True, desc="number of processes")
    baseline = File(exists=True, mandatory=True, desc="baseline T2 signal")
    gradients = File(exists=True, desc="gradients file")
    in_bvec = File(exists=True, desc="input bvecs file")
    in_bval = File(exists=True, desc="input bvals file")
    num_dirs = traits.Int(
        32,
        usedefault=True,
        desc=(
            "number of gradient directions (when table " "is automatically generated)"
        ),
    )
    bvalues = traits.List(
        traits.Int,
        value=[1000, 3000],
        usedefault=True,
        desc=("list of b-values (when table " "is automatically generated)"),
    )
    out_file = File(
        "sim_dwi.nii.gz",
        usedefault=True,
        desc="output file with fractions to be simluated",
    )
    out_mask = File(
        "sim_msk.nii.gz", usedefault=True, desc="file with the mask simulated"
    )
    out_bvec = File("bvec.sim", usedefault=True, desc="simulated b vectors")
    out_bval = File("bval.sim", usedefault=True, desc="simulated b values")
    snr = traits.Int(0, usedefault=True, desc="signal-to-noise ratio (dB)")


class SimulateMultiTensorOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc="simulated DWIs")
    out_mask = File(exists=True, desc="mask file")
    out_bvec = File(exists=True, desc="simulated b vectors")
    out_bval = File(exists=True, desc="simulated b values")


class SimulateMultiTensor(DipyBaseInterface):
    """
    Interface to MultiTensor model simulator in dipy
    http://nipy.org/dipy/examples_built/simulate_multi_tensor.html

    Example
    -------

    >>> import nipype.interfaces.dipy as dipy
    >>> sim = dipy.SimulateMultiTensor()
    >>> sim.inputs.in_dirs = ['fdir00.nii', 'fdir01.nii']
    >>> sim.inputs.in_frac = ['ffra00.nii', 'ffra01.nii']
    >>> sim.inputs.in_vfms = ['tpm_00.nii.gz', 'tpm_01.nii.gz',
    ...                       'tpm_02.nii.gz']
    >>> sim.inputs.baseline = 'b0.nii'
    >>> sim.inputs.in_bvec = 'bvecs'
    >>> sim.inputs.in_bval = 'bvals'
    >>> sim.run()                                   # doctest: +SKIP
    """

    input_spec = SimulateMultiTensorInputSpec
    output_spec = SimulateMultiTensorOutputSpec

    def _run_interface(self, runtime):
        from dipy.core.gradients import gradient_table

        # Gradient table
        if isdefined(self.inputs.in_bval) and isdefined(self.inputs.in_bvec):
            # Load the gradient strengths and directions
            bvals = np.loadtxt(self.inputs.in_bval)
            bvecs = np.loadtxt(self.inputs.in_bvec).T
            gtab = gradient_table(bvals, bvecs)
        else:
            gtab = _generate_gradients(self.inputs.num_dirs, self.inputs.bvalues)
        ndirs = len(gtab.bvals)
        np.savetxt(op.abspath(self.inputs.out_bvec), gtab.bvecs.T)
        np.savetxt(op.abspath(self.inputs.out_bval), gtab.bvals)

        # Load the baseline b0 signal
        b0_im = nb.load(self.inputs.baseline)
        hdr = b0_im.header
        shape = b0_im.shape
        aff = b0_im.affine

        # Check and load sticks and their volume fractions
        nsticks = len(self.inputs.in_dirs)
        if len(self.inputs.in_frac) != nsticks:
            raise RuntimeError(
                ("Number of sticks and their volume fractions" " must match.")
            )

        # Volume fractions of isotropic compartments
        nballs = len(self.inputs.in_vfms)
        vfs = np.squeeze(nb.concat_images(self.inputs.in_vfms).dataobj)
        if nballs == 1:
            vfs = vfs[..., np.newaxis]
        total_vf = np.sum(vfs, axis=3)

        # Generate a mask
        if isdefined(self.inputs.in_mask):
            msk = np.asanyarray(nb.load(self.inputs.in_mask).dataobj)
            msk[msk > 0.0] = 1.0
            msk[msk < 1.0] = 0.0
        else:
            msk = np.zeros(shape)
            msk[total_vf > 0.0] = 1.0

        msk = np.clip(msk, 0.0, 1.0)
        nvox = len(msk[msk > 0])

        # Fiber fractions
        ffsim = nb.concat_images(self.inputs.in_frac)
        ffs = np.nan_to_num(np.squeeze(ffsim.dataobj))  # fiber fractions
        ffs = np.clip(ffs, 0.0, 1.0)
        if nsticks == 1:
            ffs = ffs[..., np.newaxis]

        for i in range(nsticks):
            ffs[..., i] *= msk

        total_ff = np.sum(ffs, axis=3)

        # Fix incongruencies in fiber fractions
        for i in range(1, nsticks):
            if np.any(total_ff > 1.0):
                errors = np.zeros_like(total_ff)
                errors[total_ff > 1.0] = total_ff[total_ff > 1.0] - 1.0
                ffs[..., i] -= errors
                ffs[ffs < 0.0] = 0.0
            total_ff = np.sum(ffs, axis=3)

        for i in range(vfs.shape[-1]):
            vfs[..., i] -= total_ff
        vfs = np.clip(vfs, 0.0, 1.0)

        fractions = np.concatenate((ffs, vfs), axis=3)

        nb.Nifti1Image(fractions, aff, None).to_filename("fractions.nii.gz")
        nb.Nifti1Image(np.sum(fractions, axis=3), aff, None).to_filename(
            "total_vf.nii.gz"
        )

        mhdr = hdr.copy()
        mhdr.set_data_dtype(np.uint8)
        mhdr.set_xyzt_units("mm", "sec")
        nb.Nifti1Image(msk, aff, mhdr).to_filename(op.abspath(self.inputs.out_mask))

        # Initialize stack of args
        fracs = fractions[msk > 0]

        # Stack directions
        dirs = None
        for i in range(nsticks):
            f = self.inputs.in_dirs[i]
            fd = np.nan_to_num(nb.load(f).dataobj)
            w = np.linalg.norm(fd, axis=3)[..., np.newaxis]
            w[w < np.finfo(float).eps] = 1.0
            fd /= w
            if dirs is None:
                dirs = fd[msk > 0].copy()
            else:
                dirs = np.hstack((dirs, fd[msk > 0]))

        # Add random directions for isotropic components
        for d in range(nballs):
            fd = np.random.randn(nvox, 3)
            w = np.linalg.norm(fd, axis=1)
            fd[w < np.finfo(float).eps, ...] = np.array([1.0, 0.0, 0.0])
            w[w < np.finfo(float).eps] = 1.0
            fd /= w[..., np.newaxis]
            dirs = np.hstack((dirs, fd))

        sf_evals = list(self.inputs.diff_sf)
        ba_evals = list(self.inputs.diff_iso)

        mevals = [sf_evals] * nsticks + [[ba_evals[d]] * 3 for d in range(nballs)]

        b0 = b0_im.get_fdata()[msk > 0]
        args = []
        for i in range(nvox):
            args.append(
                {
                    "fractions": fracs[i, ...].tolist(),
                    "sticks": [
                        tuple(dirs[i, j : j + 3]) for j in range(nsticks + nballs)
                    ],
                    "gradients": gtab,
                    "mevals": mevals,
                    "S0": b0[i],
                    "snr": self.inputs.snr,
                }
            )

        n_proc = self.inputs.n_proc
        if n_proc == 0:
            n_proc = cpu_count()

        try:
            pool = Pool(processes=n_proc, maxtasksperchild=50)
        except TypeError:
            pool = Pool(processes=n_proc)

        # Simulate sticks using dipy
        IFLOGGER.info(
            "Starting simulation of %d voxels, %d diffusion directions.",
            len(args),
            ndirs,
        )
        result = np.array(pool.map(_compute_voxel, args))
        if np.shape(result)[1] != ndirs:
            raise RuntimeError(
                ("Computed directions do not match number" "of b-values.")
            )

        signal = np.zeros((shape[0], shape[1], shape[2], ndirs))
        signal[msk > 0] = result

        simhdr = hdr.copy()
        simhdr.set_data_dtype(np.float32)
        simhdr.set_xyzt_units("mm", "sec")
        nb.Nifti1Image(signal.astype(np.float32), aff, simhdr).to_filename(
            op.abspath(self.inputs.out_file)
        )

        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs["out_file"] = op.abspath(self.inputs.out_file)
        outputs["out_mask"] = op.abspath(self.inputs.out_mask)
        outputs["out_bvec"] = op.abspath(self.inputs.out_bvec)
        outputs["out_bval"] = op.abspath(self.inputs.out_bval)

        return outputs


def _compute_voxel(args):
    """
    Simulate DW signal for one voxel. Uses the multi-tensor model and
    three isotropic compartments.

    Apparent diffusivity tensors are taken from [Alexander2002]_
    and [Pierpaoli1996]_.

    .. [Alexander2002] Alexander et al., Detection and modeling of non-Gaussian
      apparent diffusion coefficient profiles in human brain data, MRM
      48(2):331-340, 2002, doi: `10.1002/mrm.10209
      <https://doi.org/10.1002/mrm.10209>`_.
    .. [Pierpaoli1996] Pierpaoli et al., Diffusion tensor MR imaging
      of the human brain, Radiology 201:637-648. 1996.
    """
    from dipy.sims.voxel import multi_tensor

    ffs = args["fractions"]
    gtab = args["gradients"]
    signal = np.zeros_like(gtab.bvals, dtype=np.float32)

    # Simulate dwi signal
    sf_vf = np.sum(ffs)
    if sf_vf > 0.0:
        ffs = (np.array(ffs) / sf_vf) * 100
        snr = args["snr"] if args["snr"] > 0 else None

        try:
            signal, _ = multi_tensor(
                gtab,
                args["mevals"],
                S0=args["S0"],
                angles=args["sticks"],
                fractions=ffs,
                snr=snr,
            )
        except Exception:
            pass

    return signal.tolist()


def _generate_gradients(ndirs=64, values=[1000, 3000], nb0s=1):
    """
    Automatically generate a `gradient table
    <http://nipy.org/dipy/examples_built/gradients_spheres.html#example-gradients-spheres>`_

    """
    import numpy as np
    from dipy.core.sphere import disperse_charges, Sphere, HemiSphere
    from dipy.core.gradients import gradient_table

    theta = np.pi * np.random.rand(ndirs)
    phi = 2 * np.pi * np.random.rand(ndirs)
    hsph_initial = HemiSphere(theta=theta, phi=phi)
    hsph_updated, potential = disperse_charges(hsph_initial, 5000)

    values = np.atleast_1d(values).tolist()
    vertices = hsph_updated.vertices
    bvecs = vertices.copy()
    bvals = np.ones(vertices.shape[0]) * values[0]

    for v in values[1:]:
        bvecs = np.vstack((bvecs, vertices))
        bvals = np.hstack((bvals, v * np.ones(vertices.shape[0])))

    for i in range(0, nb0s):
        bvals = bvals.tolist()
        bvals.insert(0, 0)

        bvecs = bvecs.tolist()
        bvecs.insert(0, np.zeros(3))

    return gradient_table(bvals, bvecs)



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..utilities import GeneratePurePlugMask


def test_GeneratePurePlugMask_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        inputImageModalities=dict(
            argstr="--inputImageModalities %s...",
        ),
        numberOfSubSamples=dict(
            argstr="--numberOfSubSamples %s",
            sep=",",
        ),
        outputMaskFile=dict(
            argstr="--outputMaskFile %s",
            hash_files=False,
        ),
        threshold=dict(
            argstr="--threshold %f",
        ),
    )
    inputs = GeneratePurePlugMask.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_GeneratePurePlugMask_outputs():
    output_map = dict(
        outputMaskFile=dict(
            extensions=None,
        ),
    )
    outputs = GeneratePurePlugMask.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..utils import VolumeMask


def test_VolumeMask_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        aseg=dict(
            extensions=None,
            xor=["in_aseg"],
        ),
        copy_inputs=dict(),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        in_aseg=dict(
            argstr="--aseg_name %s",
            extensions=None,
            xor=["aseg"],
        ),
        left_ribbonlabel=dict(
            argstr="--label_left_ribbon %d",
            mandatory=True,
        ),
        left_whitelabel=dict(
            argstr="--label_left_white %d",
            mandatory=True,
        ),
        lh_pial=dict(
            extensions=None,
            mandatory=True,
        ),
        lh_white=dict(
            extensions=None,
            mandatory=True,
        ),
        rh_pial=dict(
            extensions=None,
            mandatory=True,
        ),
        rh_white=dict(
            extensions=None,
            mandatory=True,
        ),
        right_ribbonlabel=dict(
            argstr="--label_right_ribbon %d",
            mandatory=True,
        ),
        right_whitelabel=dict(
            argstr="--label_right_white %d",
            mandatory=True,
        ),
        save_ribbon=dict(
            argstr="--save_ribbon",
        ),
        subject_id=dict(
            argstr="%s",
            mandatory=True,
            position=-1,
            usedefault=True,
        ),
        subjects_dir=dict(),
    )
    inputs = VolumeMask.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_VolumeMask_outputs():
    output_map = dict(
        lh_ribbon=dict(
            extensions=None,
        ),
        out_ribbon=dict(
            extensions=None,
        ),
        rh_ribbon=dict(
            extensions=None,
        ),
    )
    outputs = VolumeMask.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



import os
from pathlib import Path

from nipype.interfaces.base import (
    TraitedSpec,
    CommandLineInputSpec,
    CommandLine,
    File,
    traits,
    isdefined,
)
from nipype.utils.filemanip import split_filename


class RobexInputSpec(CommandLineInputSpec):
    in_file = File(
        desc="Input volume", exists=True, mandatory=True, position=0, argstr="%s"
    )
    out_file = File(
        desc="Output volume",
        position=1,
        argstr="%s",
        hash_files=False,
        name_template='%s_brain',
        name_source=["in_file"],
        keep_extension=True,
    )
    out_mask = File(
        desc="Output mask",
        position=2,
        argstr="%s",
        hash_files=False,
        name_template='%s_brainmask',
        name_source=["in_file"],
        keep_extension=True,
    )
    seed = traits.Int(desc="Seed for random number generator", position=3, argstr="%i")


class RobexOutputSpec(TraitedSpec):
    out_file = File(desc="Output volume")
    out_mask = File(desc="Output mask")


class RobexSegment(CommandLine):
    """

    ROBEX is an automatic whole-brain extraction tool for T1-weighted MRI data (commonly known as skull stripping).
    ROBEX aims for robust skull-stripping across datasets with no parameter settings. It fits a triangular mesh,
    constrained by a shape model, to the probabilistic output of a supervised brain boundary classifier.
    Because the shape model cannot perfectly accommodate unseen cases, a small free deformation is subsequently allowed.
    The deformation is optimized using graph cuts.
    The method ROBEX is based on was published in IEEE Transactions on Medical Imaging;
    please visit the website http://www.jeiglesias.com to download the paper.

    Examples
    --------
    >>> from nipype.interfaces.robex.preprocess import RobexSegment
    >>> robex = RobexSegment()
    >>> robex.inputs.in_file = 'structural.nii'
    >>> robex.cmdline
    'runROBEX.sh structural.nii structural_brain.nii structural_brainmask.nii'
    >>> robex.run() # doctest: +SKIP

    """

    input_spec = RobexInputSpec
    output_spec = RobexOutputSpec
    _cmd = 'runROBEX.sh'



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..utils import CatMatvec


def test_CatMatvec_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        fourxfour=dict(
            argstr="-4x4",
            xor=["matrix", "oneline"],
        ),
        in_file=dict(
            argstr="%s",
            mandatory=True,
            position=-2,
        ),
        matrix=dict(
            argstr="-MATRIX",
            xor=["oneline", "fourxfour"],
        ),
        num_threads=dict(
            nohash=True,
            usedefault=True,
        ),
        oneline=dict(
            argstr="-ONELINE",
            xor=["matrix", "fourxfour"],
        ),
        out_file=dict(
            argstr=" > %s",
            extensions=None,
            keep_extension=False,
            mandatory=True,
            name_source="in_file",
            name_template="%s_cat.aff12.1D",
            position=-1,
        ),
        outputtype=dict(),
    )
    inputs = CatMatvec.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_CatMatvec_outputs():
    output_map = dict(
        out_file=dict(
            extensions=None,
        ),
    )
    outputs = CatMatvec.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:

import numpy as np
import nibabel as nb
from nipype.testing import example_data
from ..metrics import FuzzyOverlap


def test_fuzzy_overlap(tmpdir):
    tmpdir.chdir()

    # Tests with tissue probability maps
    in_mask = example_data("tpms_msk.nii.gz")
    tpms = [example_data("tpm_%02d.nii.gz" % i) for i in range(3)]
    out = FuzzyOverlap(in_ref=tpms[0], in_tst=tpms[0]).run().outputs
    assert out.dice == 1

    out = FuzzyOverlap(in_mask=in_mask, in_ref=tpms[0], in_tst=tpms[0]).run().outputs
    assert out.dice == 1

    out = FuzzyOverlap(in_mask=in_mask, in_ref=tpms[0], in_tst=tpms[1]).run().outputs
    assert 0 < out.dice < 1

    out = FuzzyOverlap(in_ref=tpms, in_tst=tpms).run().outputs
    assert out.dice == 1.0

    out = FuzzyOverlap(in_mask=in_mask, in_ref=tpms, in_tst=tpms).run().outputs
    assert out.dice == 1.0

    # Tests with synthetic 3x3x3 images
    data = np.zeros((3, 3, 3), dtype=float)
    data[0, 0, 0] = 0.5
    data[2, 2, 2] = 0.25
    data[1, 1, 1] = 0.3
    nb.Nifti1Image(data, np.eye(4)).to_filename("test1.nii.gz")

    data = np.zeros((3, 3, 3), dtype=float)
    data[0, 0, 0] = 0.6
    data[1, 1, 1] = 0.3
    nb.Nifti1Image(data, np.eye(4)).to_filename("test2.nii.gz")

    out = FuzzyOverlap(in_ref="test1.nii.gz", in_tst="test2.nii.gz").run().outputs
    assert np.allclose(out.dice, 0.82051)

    # Just considering the mask, the central pixel
    # that raised the index now is left aside.
    data = np.zeros((3, 3, 3), dtype=int)
    data[0, 0, 0] = 1
    data[2, 2, 2] = 1
    nb.Nifti1Image(data, np.eye(4)).to_filename("mask.nii.gz")

    out = (
        FuzzyOverlap(
            in_ref="test1.nii.gz", in_tst="test2.nii.gz", in_mask="mask.nii.gz"
        )
        .run()
        .outputs
    )
    assert np.allclose(out.dice, 0.74074)



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
Package contains interfaces for using existing functionality in other packages

Requires Packages to be installed
"""
__docformat__ = "restructuredtext"

from .io import DataGrabber, DataSink, SelectFiles, BIDSDataGrabber
from .utility import IdentityInterface, Rename, Function, Select, Merge



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..gtract import gtractCostFastMarching


def test_gtractCostFastMarching_inputs():
    input_map = dict(
        anisotropyWeight=dict(
            argstr="--anisotropyWeight %f",
        ),
        args=dict(
            argstr="%s",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        inputAnisotropyVolume=dict(
            argstr="--inputAnisotropyVolume %s",
            extensions=None,
        ),
        inputStartingSeedsLabelMapVolume=dict(
            argstr="--inputStartingSeedsLabelMapVolume %s",
            extensions=None,
        ),
        inputTensorVolume=dict(
            argstr="--inputTensorVolume %s",
            extensions=None,
        ),
        numberOfThreads=dict(
            argstr="--numberOfThreads %d",
        ),
        outputCostVolume=dict(
            argstr="--outputCostVolume %s",
            hash_files=False,
        ),
        outputSpeedVolume=dict(
            argstr="--outputSpeedVolume %s",
            hash_files=False,
        ),
        seedThreshold=dict(
            argstr="--seedThreshold %f",
        ),
        startingSeedsLabel=dict(
            argstr="--startingSeedsLabel %d",
        ),
        stoppingValue=dict(
            argstr="--stoppingValue %f",
        ),
    )
    inputs = gtractCostFastMarching.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_gtractCostFastMarching_outputs():
    output_map = dict(
        outputCostVolume=dict(
            extensions=None,
        ),
        outputSpeedVolume=dict(
            extensions=None,
        ),
    )
    outputs = gtractCostFastMarching.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
"""Parallel workflow execution via OAR http://oar.imag.fr
"""
import os
import stat
from time import sleep
import subprocess
import simplejson as json

from ... import logging
from ...interfaces.base import CommandLine
from .base import SGELikeBatchManagerBase, logger

iflogger = logging.getLogger("nipype.interface")


class OARPlugin(SGELikeBatchManagerBase):
    """Execute using OAR

    The plugin_args input to run can be used to control the OAR execution.
    Currently supported options are:

    - template : template to use for batch job submission
    - oarsub_args : arguments to be prepended to the job execution
                    script in the oarsub call
    - max_jobname_len: maximum length of the job name.  Default 15.

    """

    # Addtional class variables
    _max_jobname_len = 15
    _oarsub_args = ""

    def __init__(self, **kwargs):
        template = """
# oarsub -J
        """
        self._retry_timeout = 2
        self._max_tries = 2
        self._max_jobname_length = 15
        if "plugin_args" in kwargs and kwargs["plugin_args"]:
            if "oarsub_args" in kwargs["plugin_args"]:
                self._oarsub_args = kwargs["plugin_args"]["oarsub_args"]
            if "retry_timeout" in kwargs["plugin_args"]:
                self._retry_timeout = kwargs["plugin_args"]["retry_timeout"]
            if "max_tries" in kwargs["plugin_args"]:
                self._max_tries = kwargs["plugin_args"]["max_tries"]
            if "max_jobname_len" in kwargs["plugin_args"]:
                self._max_jobname_len = kwargs["plugin_args"]["max_jobname_len"]
        super(OARPlugin, self).__init__(template, **kwargs)

    def _is_pending(self, taskid):
        #  subprocess.Popen requires taskid to be a string
        proc = subprocess.Popen(
            ["oarstat", "-J", "-s", "-j", taskid],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        o, e = proc.communicate()
        parsed_result = json.loads(o)[taskid].lower()
        is_pending = ("error" not in parsed_result) and (
            "terminated" not in parsed_result
        )
        return is_pending

    def _submit_batchtask(self, scriptfile, node):
        cmd = CommandLine(
            "oarsub",
            environ=dict(os.environ),
            resource_monitor=False,
            terminal_output="allatonce",
        )
        path = os.path.dirname(scriptfile)
        oarsubargs = ""
        if self._oarsub_args:
            oarsubargs = self._oarsub_args
        if "oarsub_args" in node.plugin_args:
            if "overwrite" in node.plugin_args and node.plugin_args["overwrite"]:
                oarsubargs = node.plugin_args["oarsub_args"]
            else:
                oarsubargs += " " + node.plugin_args["oarsub_args"]

        if node._hierarchy:
            jobname = ".".join((dict(os.environ)["LOGNAME"], node._hierarchy, node._id))
        else:
            jobname = ".".join((dict(os.environ)["LOGNAME"], node._id))
        jobnameitems = jobname.split(".")
        jobnameitems.reverse()
        jobname = ".".join(jobnameitems)
        jobname = jobname[0 : self._max_jobname_len]

        if "-O" not in oarsubargs:
            oarsubargs = "%s -O %s" % (
                oarsubargs,
                os.path.join(path, jobname + ".stdout"),
            )
        if "-E" not in oarsubargs:
            oarsubargs = "%s -E %s" % (
                oarsubargs,
                os.path.join(path, jobname + ".stderr"),
            )
        if "-J" not in oarsubargs:
            oarsubargs = "%s -J" % (oarsubargs)

        os.chmod(scriptfile, stat.S_IEXEC | stat.S_IREAD | stat.S_IWRITE)
        cmd.inputs.args = "%s -n %s -S %s" % (oarsubargs, jobname, scriptfile)

        oldlevel = iflogger.level
        iflogger.setLevel(logging.getLevelName("CRITICAL"))
        tries = 0
        while True:
            try:
                result = cmd.run()
            except Exception as e:
                if tries < self._max_tries:
                    tries += 1
                    sleep(self._retry_timeout)
                    # sleep 2 seconds and try again.
                else:
                    iflogger.setLevel(oldlevel)
                    raise RuntimeError(
                        "\n".join(
                            (
                                ("Could not submit OAR task" " for node %s") % node._id,
                                str(e),
                            )
                        )
                    )
            else:
                break
        iflogger.setLevel(oldlevel)
        # retrieve OAR taskid

        o = ""
        add = False
        for line in result.runtime.stdout.splitlines():
            if line.strip().startswith("{"):
                add = True
            if add:
                o += line + "\n"
            if line.strip().startswith("}"):
                break
        taskid = json.loads(o)["job_id"]
        self._pending[taskid] = node.output_dir()
        logger.debug("submitted OAR task: %s for node %s" % (taskid, node._id))
        return taskid



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..convert import MRTrix2TrackVis


def test_MRTrix2TrackVis_inputs():
    input_map = dict(
        image_file=dict(
            extensions=None,
        ),
        in_file=dict(
            extensions=None,
            mandatory=True,
        ),
        matrix_file=dict(
            extensions=None,
        ),
        out_filename=dict(
            extensions=None,
            genfile=True,
            usedefault=True,
        ),
        registration_image_file=dict(
            extensions=None,
        ),
    )
    inputs = MRTrix2TrackVis.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_MRTrix2TrackVis_outputs():
    output_map = dict(
        out_file=dict(
            extensions=None,
        ),
    )
    outputs = MRTrix2TrackVis.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..utils import SH2Amp


def test_SH2Amp_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        directions=dict(
            argstr="%s",
            extensions=None,
            mandatory=True,
            position=-2,
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        in_file=dict(
            argstr="%s",
            extensions=None,
            mandatory=True,
            position=-3,
        ),
        nonnegative=dict(
            argstr="-nonnegative",
        ),
        out_file=dict(
            argstr="%s",
            extensions=None,
            name_source=["in_file"],
            name_template="%s_amp.mif",
            position=-1,
            usedefault=True,
        ),
    )
    inputs = SH2Amp.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_SH2Amp_outputs():
    output_map = dict(
        out_file=dict(
            extensions=None,
        ),
    )
    outputs = SH2Amp.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..brains import BRAINSSnapShotWriter


def test_BRAINSSnapShotWriter_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        inputBinaryVolumes=dict(
            argstr="--inputBinaryVolumes %s...",
        ),
        inputPlaneDirection=dict(
            argstr="--inputPlaneDirection %s",
            sep=",",
        ),
        inputSliceToExtractInIndex=dict(
            argstr="--inputSliceToExtractInIndex %s",
            sep=",",
        ),
        inputSliceToExtractInPercent=dict(
            argstr="--inputSliceToExtractInPercent %s",
            sep=",",
        ),
        inputSliceToExtractInPhysicalPoint=dict(
            argstr="--inputSliceToExtractInPhysicalPoint %s",
            sep=",",
        ),
        inputVolumes=dict(
            argstr="--inputVolumes %s...",
        ),
        outputFilename=dict(
            argstr="--outputFilename %s",
            hash_files=False,
        ),
    )
    inputs = BRAINSSnapShotWriter.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_BRAINSSnapShotWriter_outputs():
    output_map = dict(
        outputFilename=dict(
            extensions=None,
        ),
    )
    outputs = BRAINSSnapShotWriter.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
# -*- coding: utf8 -*-
"""Autogenerated file - DO NOT EDIT
If you spot a bug, please report it on the mailing list and/or change the generator."""

from nipype.interfaces.base import (
    CommandLine,
    CommandLineInputSpec,
    SEMLikeCommandLine,
    TraitedSpec,
    File,
    Directory,
    traits,
    isdefined,
    InputMultiPath,
    OutputMultiPath,
)
import os


class BRAINSFitInputSpec(CommandLineInputSpec):
    fixedVolume = File(
        desc="The fixed image for registration by mutual information optimization.",
        exists=True,
        argstr="--fixedVolume %s",
    )
    movingVolume = File(
        desc="The moving image for registration by mutual information optimization.",
        exists=True,
        argstr="--movingVolume %s",
    )
    bsplineTransform = traits.Either(
        traits.Bool,
        File(),
        hash_files=False,
        desc="(optional) Filename to which save the estimated transform. NOTE: You must set at least one output object (either a deformed image or a transform.  NOTE: USE THIS ONLY IF THE FINAL TRANSFORM IS BSpline",
        argstr="--bsplineTransform %s",
    )
    linearTransform = traits.Either(
        traits.Bool,
        File(),
        hash_files=False,
        desc="(optional) Filename to which save the estimated transform. NOTE: You must set at least one output object (either a deformed image or a transform.  NOTE: USE THIS ONLY IF THE FINAL TRANSFORM IS ---NOT--- BSpline",
        argstr="--linearTransform %s",
    )
    outputVolume = traits.Either(
        traits.Bool,
        File(),
        hash_files=False,
        desc="(optional) Output image for registration. NOTE: You must select either the outputTransform or the outputVolume option.",
        argstr="--outputVolume %s",
    )
    initialTransform = File(
        desc="Filename of transform used to initialize the registration.  This CAN NOT be used with either CenterOfHeadLAlign, MomentsAlign, GeometryAlign, or initialTransform file.",
        exists=True,
        argstr="--initialTransform %s",
    )
    initializeTransformMode = traits.Enum(
        "Off",
        "useMomentsAlign",
        "useCenterOfHeadAlign",
        "useGeometryAlign",
        "useCenterOfROIAlign",
        desc="Determine how to initialize the transform center.  GeometryAlign on assumes that the center of the voxel lattice of the images represent similar structures.  MomentsAlign assumes that the center of mass of the images represent similar structures.  useCenterOfHeadAlign attempts to use the top of head and shape of neck to drive a center of mass estimate.  Off assumes that the physical space of the images are close, and that centering in terms of the image Origins is a good starting point.  This flag is mutually exclusive with the initialTransform flag.",
        argstr="--initializeTransformMode %s",
    )
    useRigid = traits.Bool(
        desc="Perform a rigid registration as part of the sequential registration steps.  This family of options superceeds the use of transformType if any of them are set.",
        argstr="--useRigid ",
    )
    useScaleVersor3D = traits.Bool(
        desc="Perform a ScaleVersor3D registration as part of the sequential registration steps.  This family of options superceeds the use of transformType if any of them are set.",
        argstr="--useScaleVersor3D ",
    )
    useScaleSkewVersor3D = traits.Bool(
        desc="Perform a ScaleSkewVersor3D registration as part of the sequential registration steps.  This family of options superceeds the use of transformType if any of them are set.",
        argstr="--useScaleSkewVersor3D ",
    )
    useAffine = traits.Bool(
        desc="Perform an Affine registration as part of the sequential registration steps.  This family of options superceeds the use of transformType if any of them are set.",
        argstr="--useAffine ",
    )
    useBSpline = traits.Bool(
        desc="Perform a BSpline registration as part of the sequential registration steps.  This family of options superceeds the use of transformType if any of them are set.",
        argstr="--useBSpline ",
    )
    numberOfSamples = traits.Int(
        desc="The number of voxels sampled for mutual information computation.  Increase this for a slower, more careful fit.  You can also limit the sampling focus with ROI masks and ROIAUTO mask generation.",
        argstr="--numberOfSamples %d",
    )
    splineGridSize = InputMultiPath(
        traits.Int,
        desc="The number of subdivisions of the BSpline Grid to be centered on the image space.  Each dimension must have at least 3 subdivisions for the BSpline to be correctly computed. ",
        sep=",",
        argstr="--splineGridSize %s",
    )
    numberOfIterations = InputMultiPath(
        traits.Int,
        desc="The maximum number of iterations to try before failing to converge.  Use an explicit limit like 500 or 1000 to manage risk of divergence",
        sep=",",
        argstr="--numberOfIterations %s",
    )
    maskProcessingMode = traits.Enum(
        "NOMASK",
        "ROIAUTO",
        "ROI",
        desc="What mode to use for using the masks.  If ROIAUTO is choosen, then the mask is implicitly defined using a otsu forground and hole filling algorithm. The Region Of Interest mode (choose ROI) uses the masks to define what parts of the image should be used for computing the transform.",
        argstr="--maskProcessingMode %s",
    )
    fixedBinaryVolume = File(
        desc="Fixed Image binary mask volume, ONLY FOR MANUAL ROI mode.",
        exists=True,
        argstr="--fixedBinaryVolume %s",
    )
    movingBinaryVolume = File(
        desc="Moving Image binary mask volume, ONLY FOR MANUAL ROI mode.",
        exists=True,
        argstr="--movingBinaryVolume %s",
    )
    outputFixedVolumeROI = traits.Either(
        traits.Bool,
        File(),
        hash_files=False,
        desc="The ROI automatically found in fixed image, ONLY FOR ROIAUTO mode.",
        argstr="--outputFixedVolumeROI %s",
    )
    outputMovingVolumeROI = traits.Either(
        traits.Bool,
        File(),
        hash_files=False,
        desc="The ROI automatically found in moving image, ONLY FOR ROIAUTO mode.",
        argstr="--outputMovingVolumeROI %s",
    )
    outputVolumePixelType = traits.Enum(
        "float",
        "short",
        "ushort",
        "int",
        "uint",
        "uchar",
        desc="The output image Pixel Type is the scalar datatype for representation of the Output Volume.",
        argstr="--outputVolumePixelType %s",
    )
    backgroundFillValue = traits.Float(
        desc="Background fill value for output image.",
        argstr="--backgroundFillValue %f",
    )
    maskInferiorCutOffFromCenter = traits.Float(
        desc="For use with --useCenterOfHeadAlign (and --maskProcessingMode ROIAUTO): the cut-off below the image centers, in millimeters, ",
        argstr="--maskInferiorCutOffFromCenter %f",
    )
    scaleOutputValues = traits.Bool(
        desc="If true, and the voxel values do not fit within the minimum and maximum values of the desired outputVolumePixelType, then linearly scale the min/max output image voxel values to fit within the min/max range of the outputVolumePixelType.",
        argstr="--scaleOutputValues ",
    )
    interpolationMode = traits.Enum(
        "NearestNeighbor",
        "Linear",
        "ResampleInPlace",
        "BSpline",
        "WindowedSinc",
        "Hamming",
        "Cosine",
        "Welch",
        "Lanczos",
        "Blackman",
        desc="Type of interpolation to be used when applying transform to moving volume.  Options are Linear, NearestNeighbor, BSpline, WindowedSinc, or ResampleInPlace.  The ResampleInPlace option will create an image with the same discrete voxel values and will adjust the origin and direction of the physical space interpretation.",
        argstr="--interpolationMode %s",
    )
    minimumStepLength = InputMultiPath(
        traits.Float,
        desc="Each step in the optimization takes steps at least this big.  When none are possible, registration is complete.",
        sep=",",
        argstr="--minimumStepLength %s",
    )
    translationScale = traits.Float(
        desc="How much to scale up changes in position compared to unit rotational changes in radians -- decrease this to put more rotation in the search pattern.",
        argstr="--translationScale %f",
    )
    reproportionScale = traits.Float(
        desc="ScaleVersor3D 'Scale' compensation factor.  Increase this to put more rescaling in a ScaleVersor3D or ScaleSkewVersor3D search pattern.  1.0 works well with a translationScale of 1000.0",
        argstr="--reproportionScale %f",
    )
    skewScale = traits.Float(
        desc="ScaleSkewVersor3D Skew compensation factor.  Increase this to put more skew in a ScaleSkewVersor3D search pattern.  1.0 works well with a translationScale of 1000.0",
        argstr="--skewScale %f",
    )
    maxBSplineDisplacement = traits.Float(
        desc=" Sets the maximum allowed displacements in image physical coordinates for BSpline control grid along each axis.  A value of 0.0 indicates that the problem should be unbounded.  NOTE:  This only constrains the BSpline portion, and does not limit the displacement from the associated bulk transform.  This can lead to a substantial reduction in computation time in the BSpline optimizer.,       ",
        argstr="--maxBSplineDisplacement %f",
    )
    histogramMatch = traits.Bool(
        desc="Histogram Match the input images.  This is suitable for images of the same modality that may have different absolute scales, but the same overall intensity profile. Do NOT use if registering images from different modailties.",
        argstr="--histogramMatch ",
    )
    numberOfHistogramBins = traits.Int(
        desc="The number of histogram levels", argstr="--numberOfHistogramBins %d"
    )
    numberOfMatchPoints = traits.Int(
        desc="the number of match points", argstr="--numberOfMatchPoints %d"
    )
    strippedOutputTransform = traits.Either(
        traits.Bool,
        File(),
        hash_files=False,
        desc="File name for the rigid component of the estimated affine transform. Can be used to rigidly register the moving image to the fixed image. NOTE:  This value is overwritten if either bsplineTransform or linearTransform is set.",
        argstr="--strippedOutputTransform %s",
    )
    transformType = InputMultiPath(
        traits.Str,
        desc="Specifies a list of registration types to be used.  The valid types are, Rigid, ScaleVersor3D, ScaleSkewVersor3D, Affine, and BSpline.  Specifiying more than one in a comma separated list will initialize the next stage with the previous results. If registrationClass flag is used, it overrides this parameter setting.",
        sep=",",
        argstr="--transformType %s",
    )
    outputTransform = traits.Either(
        traits.Bool,
        File(),
        hash_files=False,
        desc="(optional) Filename to which save the (optional) estimated transform. NOTE: You must select either the outputTransform or the outputVolume option.",
        argstr="--outputTransform %s",
    )
    fixedVolumeTimeIndex = traits.Int(
        desc="The index in the time series for the 3D fixed image to fit, if 4-dimensional.",
        argstr="--fixedVolumeTimeIndex %d",
    )
    movingVolumeTimeIndex = traits.Int(
        desc="The index in the time series for the 3D moving image to fit, if 4-dimensional.",
        argstr="--movingVolumeTimeIndex %d",
    )
    medianFilterSize = InputMultiPath(
        traits.Int,
        desc="The radius for the optional MedianImageFilter preprocessing in all 3 directions.",
        sep=",",
        argstr="--medianFilterSize %s",
    )
    removeIntensityOutliers = traits.Float(
        desc="The half percentage to decide outliers of image intensities. The default value is zero, which means no outlier removal. If the value of 0.005 is given, the moduel will throw away 0.005 % of both tails, so 0.01% of intensities in total would be ignored in its statistic calculation. ",
        argstr="--removeIntensityOutliers %f",
    )
    useCachingOfBSplineWeightsMode = traits.Enum(
        "ON",
        "OFF",
        desc="This is a 5x speed advantage at the expense of requiring much more memory.  Only relevant when transformType is BSpline.",
        argstr="--useCachingOfBSplineWeightsMode %s",
    )
    useExplicitPDFDerivativesMode = traits.Enum(
        "AUTO",
        "ON",
        "OFF",
        desc="Using mode AUTO means OFF for BSplineDeformableTransforms and ON for the linear transforms.  The ON alternative uses more memory to sometimes do a better job.",
        argstr="--useExplicitPDFDerivativesMode %s",
    )
    ROIAutoDilateSize = traits.Float(
        desc="This flag is only relavent when using ROIAUTO mode for initializing masks.  It defines the final dilation size to capture a bit of background outside the tissue region.  At setting of 10mm has been shown to help regularize a BSpline registration type so that there is some background constraints to match the edges of the head better.",
        argstr="--ROIAutoDilateSize %f",
    )
    ROIAutoClosingSize = traits.Float(
        desc="This flag is only relavent when using ROIAUTO mode for initializing masks.  It defines the hole closing size in mm.  It is rounded up to the nearest whole pixel size in each direction. The default is to use a closing size of 9mm.  For mouse data this value may need to be reset to 0.9 or smaller.",
        argstr="--ROIAutoClosingSize %f",
    )
    relaxationFactor = traits.Float(
        desc="Internal debugging parameter, and should probably never be used from the command line.  This will be removed in the future.",
        argstr="--relaxationFactor %f",
    )
    maximumStepLength = traits.Float(
        desc="Internal debugging parameter, and should probably never be used from the command line.  This will be removed in the future.",
        argstr="--maximumStepLength %f",
    )
    failureExitCode = traits.Int(
        desc="If the fit fails, exit with this status code.  (It can be used to force a successfult exit status of (0) if the registration fails due to reaching the maximum number of iterations.",
        argstr="--failureExitCode %d",
    )
    writeTransformOnFailure = traits.Bool(
        desc="Flag to save the final transform even if the numberOfIterations are reached without convergence. (Intended for use when --failureExitCode 0 )",
        argstr="--writeTransformOnFailure ",
    )
    numberOfThreads = traits.Int(
        desc="Explicitly specify the maximum number of threads to use. (default is auto-detected)",
        argstr="--numberOfThreads %d",
    )
    forceMINumberOfThreads = traits.Int(
        desc="Force the the maximum number of threads to use for non thread safe MI metric.  CAUTION: Inconsistent results my arise!",
        argstr="--forceMINumberOfThreads %d",
    )
    debugLevel = traits.Int(
        desc="Display debug messages, and produce debug intermediate results.  0=OFF, 1=Minimal, 10=Maximum debugging.",
        argstr="--debugLevel %d",
    )
    costFunctionConvergenceFactor = traits.Float(
        desc=" From itkLBFGSBOptimizer.h: Set/Get the CostFunctionConvergenceFactor. Algorithm terminates when the reduction in cost function is less than (factor * epsmcj) where epsmch is the machine precision. Typical values for factor: 1e+12 for low accuracy; 1e+7 for moderate accuracy and 1e+1 for extremely high accuracy.  1e+9 seems to work well.,       ",
        argstr="--costFunctionConvergenceFactor %f",
    )
    projectedGradientTolerance = traits.Float(
        desc=" From itkLBFGSBOptimizer.h: Set/Get the ProjectedGradientTolerance. Algorithm terminates when the project gradient is below the tolerance. Default lbfgsb value is 1e-5, but 1e-4 seems to work well.,       ",
        argstr="--projectedGradientTolerance %f",
    )
    gui = traits.Bool(
        desc="Display intermediate image volumes for debugging.  NOTE:  This is not part of the standard build sytem, and probably does nothing on your installation.",
        argstr="--gui ",
    )
    promptUser = traits.Bool(
        desc="Prompt the user to hit enter each time an image is sent to the DebugImageViewer",
        argstr="--promptUser ",
    )
    NEVER_USE_THIS_FLAG_IT_IS_OUTDATED_00 = traits.Bool(
        desc="DO NOT USE THIS FLAG", argstr="--NEVER_USE_THIS_FLAG_IT_IS_OUTDATED_00 "
    )
    NEVER_USE_THIS_FLAG_IT_IS_OUTDATED_01 = traits.Bool(
        desc="DO NOT USE THIS FLAG", argstr="--NEVER_USE_THIS_FLAG_IT_IS_OUTDATED_01 "
    )
    NEVER_USE_THIS_FLAG_IT_IS_OUTDATED_02 = traits.Bool(
        desc="DO NOT USE THIS FLAG", argstr="--NEVER_USE_THIS_FLAG_IT_IS_OUTDATED_02 "
    )
    permitParameterVariation = InputMultiPath(
        traits.Int,
        desc="A bit vector to permit linear transform parameters to vary under optimization.  The vector order corresponds with transform parameters, and beyond the end ones fill in as a default.  For instance, you can choose to rotate only in x (pitch) with 1,0,0;  this is mostly for expert use in turning on and off individual degrees of freedom in rotation, translation or scaling without multiplying the number of transform representations; this trick is probably meaningless when tried with the general affine transform.",
        sep=",",
        argstr="--permitParameterVariation %s",
    )
    costMetric = traits.Enum(
        "MMI",
        "MSE",
        "NC",
        "MC",
        desc="The cost metric to be used during fitting. Defaults to MMI. Options are MMI (Mattes Mutual Information), MSE (Mean Square Error), NC (Normalized Correlation), MC (Match Cardinality for binary images)",
        argstr="--costMetric %s",
    )
    writeOutputTransformInFloat = traits.Bool(
        desc="By default, the output registration transforms (either the output composite transform or each transform component) are written to the disk in double precision. If this flag is ON, the output transforms will be written in single (float) precision. It is especially important if the output transform is a displacement field transform, or it is a composite transform that includes several displacement fields.",
        argstr="--writeOutputTransformInFloat ",
    )


class BRAINSFitOutputSpec(TraitedSpec):
    bsplineTransform = File(
        desc="(optional) Filename to which save the estimated transform. NOTE: You must set at least one output object (either a deformed image or a transform.  NOTE: USE THIS ONLY IF THE FINAL TRANSFORM IS BSpline",
        exists=True,
    )
    linearTransform = File(
        desc="(optional) Filename to which save the estimated transform. NOTE: You must set at least one output object (either a deformed image or a transform.  NOTE: USE THIS ONLY IF THE FINAL TRANSFORM IS ---NOT--- BSpline",
        exists=True,
    )
    outputVolume = File(
        desc="(optional) Output image for registration. NOTE: You must select either the outputTransform or the outputVolume option.",
        exists=True,
    )
    outputFixedVolumeROI = File(
        desc="The ROI automatically found in fixed image, ONLY FOR ROIAUTO mode.",
        exists=True,
    )
    outputMovingVolumeROI = File(
        desc="The ROI automatically found in moving image, ONLY FOR ROIAUTO mode.",
        exists=True,
    )
    strippedOutputTransform = File(
        desc="File name for the rigid component of the estimated affine transform. Can be used to rigidly register the moving image to the fixed image. NOTE:  This value is overwritten if either bsplineTransform or linearTransform is set.",
        exists=True,
    )
    outputTransform = File(
        desc="(optional) Filename to which save the (optional) estimated transform. NOTE: You must select either the outputTransform or the outputVolume option.",
        exists=True,
    )


class BRAINSFit(SEMLikeCommandLine):
    """title: General Registration (BRAINS)

    category: Registration

    description: Register a three-dimensional volume to a reference volume (Mattes Mutual Information by default). Described in BRAINSFit: Mutual Information Registrations of Whole-Brain 3D Images, Using the Insight Toolkit, Johnson H.J., Harris G., Williams K., The Insight Journal, 2007. http://hdl.handle.net/1926/1291

    version: 3.0.0

    documentation-url: http://wiki.slicer.org/slicerWiki/index.php/Modules:BRAINSFit

    license: https://www.nitrc.org/svn/brains/BuildScripts/trunk/License.txt

    contributor: Hans J. Johnson, hans-johnson -at- uiowa.edu, http://wwww.psychiatry.uiowa.edu

    acknowledgements: Hans Johnson(1,3,4); Kent Williams(1); Gregory Harris(1), Vincent Magnotta(1,2,3);  Andriy Fedorov(5) 1=University of Iowa Department of Psychiatry, 2=University of Iowa Department of Radiology, 3=University of Iowa Department of Biomedical Engineering, 4=University of Iowa Department of Electrical and Computer Engineering, 5=Surgical Planning Lab, Harvard
    """

    input_spec = BRAINSFitInputSpec
    output_spec = BRAINSFitOutputSpec
    _cmd = "BRAINSFit "
    _outputs_filenames = {
        "outputVolume": "outputVolume.nii",
        "bsplineTransform": "bsplineTransform.mat",
        "outputTransform": "outputTransform.mat",
        "outputFixedVolumeROI": "outputFixedVolumeROI.nii",
        "strippedOutputTransform": "strippedOutputTransform.mat",
        "outputMovingVolumeROI": "outputMovingVolumeROI.nii",
        "linearTransform": "linearTransform.mat",
    }



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
The fusion module provides higher-level interfaces to some of the operations
that can be performed with the seg_DetectLesions command-line program.
"""

import warnings

from ..base import TraitedSpec, File, traits, CommandLineInputSpec
from .base import NiftySegCommand
from ..niftyreg.base import get_custom_path

warn = warnings.warn
warnings.filterwarnings("always", category=UserWarning)


class PatchMatchInputSpec(CommandLineInputSpec):
    """Input Spec for PatchMatch."""

    # Mandatory input arguments
    in_file = File(
        argstr="-i %s",
        exists=True,
        mandatory=True,
        desc="Input image to segment",
        position=1,
    )

    mask_file = File(
        argstr="-m %s",
        exists=True,
        mandatory=True,
        desc="Input mask for the area where applies PatchMatch",
        position=2,
    )

    database_file = File(
        argstr="-db %s",
        exists=True,
        mandatory=True,
        desc="Database with the segmentations",
        position=3,
    )

    # Output file name
    out_file = File(
        name_source=["in_file"],
        name_template="%s_pm.nii.gz",
        desc="The output filename of the patchmatch results",
        argstr="-o %s",
        position=4,
    )

    # Optional arguments
    patch_size = traits.Int(desc="Patch size, #voxels", argstr="-size %i")

    desc = "Constrained search area size, number of times bigger than the \
patchsize"

    cs_size = traits.Int(desc=desc, argstr="-cs %i")

    match_num = traits.Int(desc="Number of better matching", argstr="-match %i")

    pm_num = traits.Int(desc="Number of patchmatch executions", argstr="-pm %i")

    desc = "Number of iterations for the patchmatch algorithm"
    it_num = traits.Int(desc=desc, argstr="-it %i")


class PatchMatchOutputSpec(TraitedSpec):
    """OutputSpec for PatchMatch."""

    out_file = File(desc="Output segmentation")


class PatchMatch(NiftySegCommand):
    """Interface for executable seg_PatchMatch from NiftySeg platform.

    The database file is a text file and in each line we have a template
    file, a mask with the search region to consider and a file with the
    label to propagate.

    Input image, input mask, template images from database and masks from
    database must have the same 4D resolution (same number of XxYxZ voxels,
    modalities and/or time-points).
    Label files from database must have the same 3D resolution
    (XxYxZ voxels) than input image but can have different number of
    volumes than the input image allowing to propagate multiple labels
    in the same execution.

    `Source code <http://cmictig.cs.ucl.ac.uk/wiki/index.php/NiftySeg>`_ |
    `Documentation <http://cmictig.cs.ucl.ac.uk/wiki/index.php/NiftySeg_documentation>`_

    Examples
    --------
    >>> from nipype.interfaces import niftyseg
    >>> node = niftyseg.PatchMatch()
    >>> node.inputs.in_file = 'im1.nii'
    >>> node.inputs.mask_file = 'im2.nii'
    >>> node.inputs.database_file = 'db.xml'
    >>> node.cmdline
    'seg_PatchMatch -i im1.nii -m im2.nii -db db.xml -o im1_pm.nii.gz'

    """

    _cmd = get_custom_path("seg_PatchMatch", env_dir="NIFTYSEGDIR")
    input_spec = PatchMatchInputSpec
    output_spec = PatchMatchOutputSpec
    _suffix = "_pm"



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..icc import ICC


def test_ICC_inputs():
    input_map = dict(
        mask=dict(
            extensions=None,
            mandatory=True,
        ),
        subjects_sessions=dict(
            mandatory=True,
        ),
    )
    inputs = ICC.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_ICC_outputs():
    output_map = dict(
        icc_map=dict(
            extensions=None,
        ),
        session_var_map=dict(
            extensions=None,
        ),
        subject_var_map=dict(
            extensions=None,
        ),
    )
    outputs = ICC.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



"""
DTI-TK is a spatial normalization and atlas construction toolkit for DTI.

Interfaces for the `Diffusion Tensor Imaging Toolkit (DTI-TK)
<http://dti-tk.sourceforge.net/pmwiki/pmwiki.php>`_ command line tools.

"""
from .registration import (
    Rigid,
    Affine,
    Diffeo,
    ComposeXfm,
    DiffeoSymTensor3DVol,
    AffSymTensor3DVol,
    AffScalarVol,
    DiffeoScalarVol,
)
from .utils import (
    TVAdjustVoxSp,
    SVAdjustVoxSp,
    TVResample,
    SVResample,
    TVtool,
    BinThresh,
)



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""MeshFix corrects topological errors in polygonal meshes."""

import os.path as op
from ..utils.filemanip import split_filename
from .base import (
    CommandLine,
    CommandLineInputSpec,
    traits,
    TraitedSpec,
    isdefined,
    File,
)


class MeshFixInputSpec(CommandLineInputSpec):
    number_of_biggest_shells = traits.Int(
        argstr="--shells %d", desc="Only the N biggest shells are kept"
    )

    epsilon_angle = traits.Range(
        argstr="-a %f",
        low=0.0,
        high=2.0,
        desc="Epsilon angle in degrees (must be between 0 and 2)",
    )

    join_overlapping_largest_components = traits.Bool(
        argstr="-j",
        xor=["join_closest_components"],
        desc="Join 2 biggest components if they overlap, remove the rest.",
    )

    join_closest_components = traits.Bool(
        argstr="-jc",
        xor=["join_closest_components"],
        desc="Join the closest pair of components.",
    )

    quiet_mode = traits.Bool(
        argstr="-q", desc="Quiet mode, don't write much to stdout."
    )

    dont_clean = traits.Bool(argstr="--no-clean", desc="Don't Clean")

    save_as_stl = traits.Bool(
        xor=["save_as_vrml", "save_as_freesurfer_mesh"],
        argstr="--stl",
        desc="Result is saved in stereolithographic format (.stl)",
    )
    save_as_vrml = traits.Bool(
        argstr="--wrl",
        xor=["save_as_stl", "save_as_freesurfer_mesh"],
        desc="Result is saved in VRML1.0 format (.wrl)",
    )
    save_as_freesurfer_mesh = traits.Bool(
        argstr="--fsmesh",
        xor=["save_as_vrml", "save_as_stl"],
        desc="Result is saved in freesurfer mesh format",
    )

    remove_handles = traits.Bool(argstr="--remove-handles", desc="Remove handles")

    uniform_remeshing_steps = traits.Int(
        argstr="-u %d",
        requires=["uniform_remeshing_vertices"],
        desc="Number of steps for uniform remeshing of the whole mesh",
    )

    uniform_remeshing_vertices = traits.Int(
        argstr="--vertices %d",
        requires=["uniform_remeshing_steps"],
        desc="Constrains the number of vertices."
        "Must be used with uniform_remeshing_steps",
    )

    laplacian_smoothing_steps = traits.Int(
        argstr="--smooth %d", desc="The number of laplacian smoothing steps to apply"
    )

    x_shift = traits.Int(
        argstr="--smooth %d",
        desc="Shifts the coordinates of the vertices when saving. Output must be in FreeSurfer format",
    )

    # Cutting, decoupling, dilation
    cut_outer = traits.Int(
        argstr="--cut-outer %d",
        desc="Remove triangles of 1st that are outside of the 2nd shell.",
    )
    cut_inner = traits.Int(
        argstr="--cut-inner %d",
        desc="Remove triangles of 1st that are inside of the 2nd shell. Dilate 2nd by N; Fill holes and keep only 1st afterwards.",
    )
    decouple_inin = traits.Int(
        argstr="--decouple-inin %d",
        desc="Treat 1st file as inner, 2nd file as outer component."
        "Resolve overlaps by moving inners triangles inwards. Constrain the min distance between the components > d.",
    )
    decouple_outin = traits.Int(
        argstr="--decouple-outin %d",
        desc="Treat 1st file as outer, 2nd file as inner component."
        "Resolve overlaps by moving outers triangles inwards. Constrain the min distance between the components > d.",
    )
    decouple_outout = traits.Int(
        argstr="--decouple-outout %d",
        desc="Treat 1st file as outer, 2nd file as inner component."
        "Resolve overlaps by moving outers triangles outwards. Constrain the min distance between the components > d.",
    )

    finetuning_inwards = traits.Bool(
        argstr="--fineTuneIn ",
        requires=["finetuning_distance", "finetuning_substeps"],
        position=-3,
        desc="Used to fine-tune the minimal distance between surfaces.",
    )
    finetuning_outwards = traits.Bool(
        argstr="--fineTuneOut ",
        requires=["finetuning_distance", "finetuning_substeps"],
        position=-3,
        xor=["finetuning_inwards"],
        desc="Similar to finetuning_inwards, but ensures minimal distance in the other direction",
    )
    finetuning_distance = traits.Float(
        argstr="%f",
        requires=["finetuning_substeps"],
        position=-2,
        desc="Used to fine-tune the minimal distance between surfaces."
        "A minimal distance d is ensured, and reached in n substeps. When using the surfaces for subsequent volume meshing by gmsh, this step prevent too flat tetrahedra2)",
    )
    finetuning_substeps = traits.Int(
        argstr="%d",
        requires=["finetuning_distance"],
        position=-1,
        desc="Used to fine-tune the minimal distance between surfaces."
        "A minimal distance d is ensured, and reached in n substeps. When using the surfaces for subsequent volume meshing by gmsh, this step prevent too flat tetrahedra2)",
    )

    dilation = traits.Int(
        argstr="--dilate %d", desc="Dilate the surface by d. d < 0 means shrinking."
    )
    set_intersections_to_one = traits.Bool(
        argstr="--intersect",
        desc="If the mesh contains intersections, return value = 1."
        "If saved in gmsh format, intersections will be highlighted.",
    )

    in_file1 = File(exists=True, argstr="%s", position=1, mandatory=True)
    in_file2 = File(exists=True, argstr="%s", position=2)
    output_type = traits.Enum(
        "off",
        ["stl", "msh", "wrl", "vrml", "fs", "off"],
        usedefault=True,
        desc="The output type to save the file as.",
    )
    out_filename = File(
        genfile=True, argstr="-o %s", desc="The output filename for the fixed mesh file"
    )


class MeshFixOutputSpec(TraitedSpec):
    mesh_file = File(exists=True, desc="The output mesh file")


class MeshFix(CommandLine):
    """
    MeshFix v1.2-alpha - by Marco Attene, Mirko Windhoff, Axel Thielscher.

    .. seealso::

        http://jmeshlib.sourceforge.net
            Sourceforge page

        http://simnibs.de/installation/meshfixandgetfem
            Ubuntu installation instructions

    If MeshFix is used for research purposes, please cite the following paper:
    M. Attene - A lightweight approach to repairing digitized polygon meshes.
    The Visual Computer, 2010. (c) Springer.

    Accepted input formats are OFF, PLY and STL.
    Other formats (like .msh for gmsh) are supported only partially.

    Example
    -------

    >>> import nipype.interfaces.meshfix as mf
    >>> fix = mf.MeshFix()
    >>> fix.inputs.in_file1 = 'lh-pial.stl'
    >>> fix.inputs.in_file2 = 'rh-pial.stl'
    >>> fix.run()                                    # doctest: +SKIP
    >>> fix.cmdline
    'meshfix lh-pial.stl rh-pial.stl -o lh-pial_fixed.off'
    """

    _cmd = "meshfix"
    input_spec = MeshFixInputSpec
    output_spec = MeshFixOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        if isdefined(self.inputs.out_filename):
            path, name, ext = split_filename(self.inputs.out_filename)
            ext = ext.replace(".", "")
            out_types = ["stl", "msh", "wrl", "vrml", "fs", "off"]
            # Make sure that the output filename uses one of the possible file types
            if any(ext == out_type.lower() for out_type in out_types):
                outputs["mesh_file"] = op.abspath(self.inputs.out_filename)
            else:
                outputs["mesh_file"] = op.abspath(name + "." + self.inputs.output_type)
        else:
            outputs["mesh_file"] = op.abspath(self._gen_outfilename())
        return outputs

    def _gen_filename(self, name):
        if name == "out_filename":
            return self._gen_outfilename()
        else:
            return None

    def _gen_outfilename(self):
        _, name, _ = split_filename(self.inputs.in_file1)
        if self.inputs.save_as_freesurfer_mesh or self.inputs.output_type == "fs":
            self.inputs.output_type = "fs"
            self.inputs.save_as_freesurfer_mesh = True
        if self.inputs.save_as_stl or self.inputs.output_type == "stl":
            self.inputs.output_type = "stl"
            self.inputs.save_as_stl = True
        if self.inputs.save_as_vrml or self.inputs.output_type == "vrml":
            self.inputs.output_type = "vrml"
            self.inputs.save_as_vrml = True
        return name + "_fixed." + self.inputs.output_type
